
/*
external interface for bsde.
*/

#include "bsde_internal.h"

static BSDE_World *bsde_worlds[16];
static int bsde_world;

static int *bsde_build_array=NULL;
static double *bsde_vector_array=NULL;
static double *bsde_scalar_array=NULL;
static int bsde_build_level=0;
static int bsde_build_array_pos=0;
static int bsde_vector_array_pos=0;
static int bsde_scalar_array_pos=0;
static int bsde_build_toplevel=0;

static int bsde_seeded=0;	//set if seeds were given for the rng

char *bsdeGetString(int str)
{
	char *s;

	switch(str)
	{
	case BSDE_NULL: s=""; break;
	case BSDE_CREATOR: s="Brendan G Bohannon"; break;
	case BSDE_ENGINE: s="BGB-BSDE"; break;
	case BSDE_ENGINE_VER: s="0.02 " __DATE__; break;
	case BSDE_VERSION: s="0.02"; break;
	case BSDE_EXTENSIONS: s=""; break;
	default: s=NULL; break;
	}
	return(s);
}

void bsdeSRand(bsdeUInt seed)
{
	BSDE_SRand(seed);
	BSDE_Seed21();
	bsde_seeded=1;
}

void bsdeSRand2(bsdeUInt seed1, bsdeUInt seed2)
{
	BSDE_SRand2(seed1, seed2);
	BSDE_Seed21();
	bsde_seeded=1;
}

int bsdeNewWorld()
{
	int i;

	if(!bsde_seeded)
		bsdeSRand2(123456789, 987654321);

	for(i=0; i<16; i++)
		if(!bsde_worlds[i])
	{
		bsde_worlds[i]=BSDE_NewWorld();
		bsde_world=i;
		return(i);
	}
	return(-1);
}

void bsdeBindWorld(int i)
{
	bsde_world=i;
}

void bsdeStepWorld(double dt)
{
	BSDE_StepWorld(bsde_worlds[bsde_world], dt);
}


double bsdeGetWorldAttrF(int var)
{
	double f;

	f=BSDE_GetWorldAttrF(bsde_worlds[bsde_world], var);
	return(f);
}

void bsdeSetWorldAttrF(int var, double val)
{
	BSDE_SetWorldAttrF(bsde_worlds[bsde_world], var, val);
}

int bsdeGetWorldAttrI(int var)
{
	int i;
	i=BSDE_GetWorldAttrI(bsde_worlds[bsde_world], var);
	return(i);
}

void bsdeSetWorldAttrI(int var, int val)
{
	BSDE_SetWorldAttrI(bsde_worlds[bsde_world], var, val);
}

void bsdeGetWorldAttrFV(int var, float *val, int n)
{
	BSDE_GetWorldAttrFV(bsde_worlds[bsde_world], var, val, n);
}

void bsdeSetWorldAttrFV(int var, float *val, int n)
{
	BSDE_SetWorldAttrFV(bsde_worlds[bsde_world], var, val, n);
}

void bsdeGetWorldAttrDV(int var, double *val, int n)
{
	BSDE_GetWorldAttrDV(bsde_worlds[bsde_world], var, val, n);
}

void bsdeSetWorldAttrDV(int var, double *val, int n)
{
	BSDE_SetWorldAttrDV(bsde_worlds[bsde_world], var, val, n);
}
int *bsdeGetTriggerArray()
{
	return(BSDE_GetTriggerArray(bsde_worlds[bsde_world]));
}

int bsdeGetTriggerArrayPos()
{
	return(BSDE_GetTriggerArrayPos(bsde_worlds[bsde_world]));
}

void bsdeResetTriggerArray()
{
	BSDE_ResetTriggerArray(bsde_worlds[bsde_world]);
}

void bsdeSetTriggerArray(int *arr, int num)
{
	BSDE_SetTriggerArray(bsde_worlds[bsde_world], arr, num);
}


int bsdeNewAABB(double mx, double my, double mz,
	double nx, double ny, double nz)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewAABB(mx, my, mz, nx, ny, nz);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewOBB(double mx, double my, double mz,
	double nx, double ny, double nz)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewOBB(mx, my, mz, nx, ny, nz);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewSphere(double radius)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewSphere(radius);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewCylinder(double height, double radius)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewCylinder(height, radius);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewCappedCylinder(double height, double radius)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewCappedCylinder(height, radius);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewPlane(double x, double y, double z, double d)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewPlane(x, y, z, d);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewHullF(float *norms, int n)
{
	BSDE_Solid *tmp;
	double *np;
	int i;

	np=malloc(4*n*sizeof(double));
	for(i=0; i<(4*n); i++)np[i]=norms[i];
	tmp=BSDE_NewHull(np, n);
	free(np);

	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewHullD(double *norms, int n)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewHull(norms, n);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewTriMeshF(float *vecs, int *tris, int nv, int nt)
{
	BSDE_Solid *tmp;
	double *np;
	int i;

	np=malloc(3*nv*sizeof(double));
	for(i=0; i<(3*nv); i++)np[i]=vecs[i];
	tmp=BSDE_NewTriMesh(np, tris, nv, nt);
	free(np);

	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

int bsdeNewTriMeshD(double *vecs, int *tris, int nv, int nt)
{
	BSDE_Solid *tmp;
	int i;

	tmp=BSDE_NewTriMesh(vecs, tris, nv, nt);
	i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}


void bsdeDestroy(int obj)
{
	BSDE_Destroy(bsde_worlds[bsde_world]->solids[obj]);
	bsde_worlds[bsde_world]->solids[obj]=NULL;
}

void bsdeHide(int obj)
{
	bsde_worlds[bsde_world]->solids[obj]->moveflags|=
		BSDE_MOVEFL_HIDDEN;
}

void bsdeUnhide(int obj)
{
	bsde_worlds[bsde_world]->solids[obj]->moveflags&=
		~BSDE_MOVEFL_HIDDEN;
}

double bsdeGetAttrF(int obj, int var)
{
	double f;

	f=BSDE_GetAttrF(bsde_worlds[bsde_world]->solids[obj], var);
	return(f);
}

void bsdeSetAttrF(int obj, int var, double val)
{
	BSDE_SetAttrF(bsde_worlds[bsde_world]->solids[obj], var, val);
}

int bsdeGetAttrI(int obj, int var)
{
	int i;
	i=BSDE_GetAttrI(bsde_worlds[bsde_world]->solids[obj], var);
	return(i);
}

void bsdeSetAttrI(int obj, int var, int val)
{
	BSDE_SetAttrI(bsde_worlds[bsde_world]->solids[obj], var, val);
}

void bsdeGetAttrFV(int obj, int var, float *val, int n)
{
	BSDE_GetAttrFV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}

void bsdeSetAttrFV(int obj, int var, float *val, int n)
{
	BSDE_SetAttrFV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}

void bsdeGetAttrDV(int obj, int var, double *val, int n)
{
	BSDE_GetAttrDV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}

void bsdeSetAttrDV(int obj, int var, double *val, int n)
{
	BSDE_SetAttrDV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}

void bsdeGetAttrIV(int obj, int var, int *val, int n)
{
	BSDE_GetAttrIV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}

void bsdeSetAttrIV(int obj, int var, int *val, int n)
{
	BSDE_SetAttrIV(bsde_worlds[bsde_world]->solids[obj], var, val, n);
}


int bsdeNewJoint(int type)
{
	BSDE_Joint *tmp;
	int i;

	tmp=BSDE_NewJoint(type);
	i=BSDE_AddJointWorld(bsde_worlds[bsde_world], tmp);
	return(i);
}

void bsdeBindJointSolids(int obj, int obja, int objb)
{
	BSDE_BindJointSolids(bsde_worlds[bsde_world]->joints[obj],
		bsde_worlds[bsde_world]->solids[obja],
		bsde_worlds[bsde_world]->solids[objb]);
}

double bsdeGetJointAttrF(int obj, int var)
{
	double f;

	f=BSDE_GetJointAttrF(bsde_worlds[bsde_world]->joints[obj], var);
	return(f);
}

void bsdeSetJointAttrF(int obj, int var, double val)
{
	BSDE_SetJointAttrF(bsde_worlds[bsde_world]->joints[obj], var, val);
}

int bsdeGetJointAttrI(int obj, int var)
{
	int i;
	i=BSDE_GetJointAttrI(bsde_worlds[bsde_world]->joints[obj], var);
	return(i);
}

void bsdeSetJointAttrI(int obj, int var, int val)
{
	BSDE_SetJointAttrI(bsde_worlds[bsde_world]->joints[obj], var, val);
}

void bsdeGetJointAttrFV(int obj, int var, float *val, int n)
{
	BSDE_GetJointAttrFV(bsde_worlds[bsde_world]->joints[obj],
		var, val, n);
}

void bsdeSetJointAttrFV(int obj, int var, float *val, int n)
{
	BSDE_SetJointAttrFV(bsde_worlds[bsde_world]->joints[obj],
		var, val, n);
}

void bsdeGetJointAttrDV(int obj, int var, double *val, int n)
{
	BSDE_GetJointAttrDV(bsde_worlds[bsde_world]->joints[obj],
		var, val, n);
}

void bsdeSetJointAttrDV(int obj, int var, double *val, int n)
{
	BSDE_SetJointAttrDV(bsde_worlds[bsde_world]->joints[obj],
		var, val, n);
}


void bsdeBegin(int type)
{
	if(!bsde_build_array)
	{
		bsde_build_array=malloc((1<<18)*sizeof(int));
		bsde_vector_array=malloc(65536*4*sizeof(double));
		bsde_scalar_array=malloc(65536*sizeof(double));
	}

	if(!bsde_build_level)
	{
		bsde_build_array_pos=0;
		bsde_vector_array_pos=0;
		bsde_scalar_array_pos=0;
		bsde_build_toplevel=type;
	}else
	{
		bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_BEGIN;
		bsde_build_array[bsde_build_array_pos++]=type;
	}
	bsde_build_level++;
}

int bsdeEnd()
{
	BSDE_Solid *tmp;
	double *varr, *sarr;
	int *arr;
	int i;

	if(bsde_build_level<=0)return(-1);
	bsde_build_level--;

	if(bsde_build_level)
	{
		bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_END;
	}else
	{
		//encoded so end is still obvious to decoder
		bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_END;

		arr=bsde_build_array;
		varr=bsde_vector_array;
		sarr=bsde_scalar_array;
		tmp=BSDE_NewSolidArray(bsde_build_toplevel,
			&arr, &varr, &sarr);
		i=BSDE_AddSolidWorld(bsde_worlds[bsde_world], tmp);
		return(i);
	}
	return(0);
}

void bsdeVertex3f(double x, double y, double z)
{
	bsde_vector_array[bsde_vector_array_pos*4+0]=x;
	bsde_vector_array[bsde_vector_array_pos*4+1]=y;
	bsde_vector_array[bsde_vector_array_pos*4+2]=z;
	bsde_vector_array[bsde_vector_array_pos*4+3]=0;
	bsde_vector_array_pos++;
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_VERTEX;
}

void bsdeVertex4f(double x, double y, double z, double w)
{
	bsde_vector_array[bsde_vector_array_pos*4+0]=x;
	bsde_vector_array[bsde_vector_array_pos*4+1]=y;
	bsde_vector_array[bsde_vector_array_pos*4+2]=z;
	bsde_vector_array[bsde_vector_array_pos*4+3]=w;
	bsde_vector_array_pos++;
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_VERTEX;
}

void bsdeVertex3d(double x, double y, double z) { bsdeVertex3f(x, y, z); }
void bsdeVertex4d(double x, double y, double z, double w)
	{ bsdeVertex4f(x, y, z, w); }
void bsdeVertex3fv(float *v) { bsdeVertex3f(v[0], v[1], v[2]); }
void bsdeVertex4fv(float *v) { bsdeVertex4f(v[0], v[1], v[2], v[3]); }
void bsdeVertex3dv(double *v) { bsdeVertex3f(v[0], v[1], v[2]); }
void bsdeVertex4dv(double *v) { bsdeVertex4f(v[0], v[1], v[2], v[3]); }

void bsdeNormal3f(double x, double y, double z)
{
	bsde_vector_array[bsde_vector_array_pos*4+0]=x;
	bsde_vector_array[bsde_vector_array_pos*4+1]=y;
	bsde_vector_array[bsde_vector_array_pos*4+2]=z;
	bsde_vector_array[bsde_vector_array_pos*4+3]=0;
	bsde_vector_array_pos++;
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_NORMAL;
}

void bsdeNormal4f(double x, double y, double z, double w)
{
	bsde_vector_array[bsde_vector_array_pos*4+0]=x;
	bsde_vector_array[bsde_vector_array_pos*4+1]=y;
	bsde_vector_array[bsde_vector_array_pos*4+2]=z;
	bsde_vector_array[bsde_vector_array_pos*4+3]=w;
	bsde_vector_array_pos++;
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_NORMAL;
}

void bsdeNormal3d(double x, double y, double z) { bsdeNormal3f(x, y, z); }
void bsdeNormal4d(double x, double y, double z, double w)
	{ bsdeNormal4f(x, y, z, w); }
void bsdeNormal3fv(float *v) { bsdeNormal3f(v[0], v[1], v[2]); }
void bsdeNormal4fv(float *v) { bsdeNormal4f(v[0], v[1], v[2], v[3]); }
void bsdeNormal3dv(double *v) { bsdeNormal3f(v[0], v[1], v[2]); }
void bsdeNormal4dv(double *v) { bsdeNormal4f(v[0], v[1], v[2], v[3]); }

void bsdeAttrI(int var, int val)
{
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_ATTR_I;
	bsde_build_array[bsde_build_array_pos++]=var;
	bsde_build_array[bsde_build_array_pos++]=val;
}

void bsdeAttrF(int var, double val)
{
	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_ATTR_F;
	bsde_build_array[bsde_build_array_pos++]=var;
	bsde_scalar_array[bsde_scalar_array_pos++]=val;
}

void bsdeAttrIV(int var, int *val, int num)
{
	int i;

	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_ATTR_IV;
	bsde_build_array[bsde_build_array_pos++]=var;
	bsde_build_array[bsde_build_array_pos++]=num;

	for(i=0; i<num; i++)
		bsde_build_array[bsde_build_array_pos++]=val[i];
}

void bsdeAttrFV(int var, float *val, int num)
{
	int i;

	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_ATTR_FV;
	bsde_build_array[bsde_build_array_pos++]=var;
	bsde_build_array[bsde_build_array_pos++]=num;

	for(i=0; i<num; i++)
		bsde_scalar_array[bsde_scalar_array_pos++]=val[i];
}

void bsdeAttrDV(int var, double *val, int num)
{
	int i;

	bsde_build_array[bsde_build_array_pos++]=BSDE_TAG_ATTR_FV;
	bsde_build_array[bsde_build_array_pos++]=var;
	bsde_build_array[bsde_build_array_pos++]=num;

	for(i=0; i<num; i++)
		bsde_scalar_array[bsde_scalar_array_pos++]=val[i];
}
