/*
internal and external constants.
*/

#ifndef BSDE_CONST_H
#define BSDE_CONST_H

//0..255: reserved for small context-specific integers
#define BSDE_NULL		0x0000
#define BSDE_FALSE		0x0000
#define BSDE_TRUE		0x0001
#define BSDE_ZERO		0x0000
#define BSDE_ONE		0x0001
#define BSDE_0			0x0000
#define BSDE_1			0x0001

#define BSDE_CREATOR		0x0001
#define BSDE_ENGINE		0x0002
#define BSDE_ENGINE_VER		0x0003
#define BSDE_VERSION		0x0004
#define BSDE_EXTENSIONS		0x0005

//vector properties
#define BSDE_ORIGIN		0x0100
#define BSDE_VELOCITY		0x0101
#define BSDE_TORQUE		0x0102
#define BSDE_ANGULAR_VELOCITY	0x0103
#define BSDE_ROTATION		0x0104	//rotation quaternion
#define BSDE_MINS		0x0105
#define BSDE_MAXS		0x0106
#define BSDE_MASS_OFFSET	0x0107
#define BSDE_INERTIA_TENSOR	0x0108
#define BSDE_DIRECTION		0x0109
#define BSDE_ORIGIN_A		0x010A	//origin as per A
#define BSDE_ORIGIN_B		0x010B	//origin as per B
#define BSDE_ORIGIN_RA		0x010C	//local origin as per A
#define BSDE_ORIGIN_RB		0x010D	//local origin as per B
#define BSDE_DIRECTION_A	0x010E	//direction on A
#define BSDE_DIRECTION_B	0x010F	//direction on B
#define BSDE_DIRECTION_RA	0x0110	//direction in A's local space
#define BSDE_DIRECTION_RB	0x0111	//direction in B's local space
#define BSDE_ROTATION_MATRIX	0x0112	//3x3 rotation matrix
#define BSDE_NORMAL		0x0113
#define BSDE_VERTICES		0x0114	//solid vertices
#define BSDE_CONTACT_VELOCITY	0x0115	//velocity at contacts
#define BSDE_ACCELERATION	0x0116	//linear acceleration
#define BSDE_LOCAL_THRUST	0x0117
#define BSDE_ORIGIN_0		0x0100	//base origin
#define BSDE_ORIGIN_1		0x0118	//2nd origin (joints)
#define BSDE_ORIGIN_2		0x0119	//3rd origin (joints)
#define BSDE_ORIGIN_3		0x011A	//4th origin (joints)
#define BSDE_DIRECTION_0	0x0109	//first direction
#define BSDE_DIRECTION_1	0x011B	//second direction
#define BSDE_ORIGIN_123		0x011C	//set base and generate others

//float properties
#define BSDE_HEIGHT		0x0120
#define BSDE_RADIUS		0x0121
#define BSDE_MASS		0x0122
#define BSDE_VOLUME		0x0123
#define BSDE_DENSITY		0x0124
#define BSDE_ELASTIC		0x0125
#define BSDE_FRICTION		0x0126
#define BSDE_VISCOSITY		0x0127
#define BSDE_CONTACT_POWER	0x0128
#define BSDE_STRENGTH		0x0129
#define BSDE_STRENGTH_0		0x0129
#define BSDE_STRENGTH_1		0x012A
#define BSDE_STRENGTH_2		0x012B
#define BSDE_STRENGTH_3		0x012C

#define BSDE_CONSTANT_ATTENUATION	0x012D
#define BSDE_LINEAR_ATTENUATION		0x012E
#define BSDE_QUADRATIC_ATTENUATION	0x012F

//integer/boolean properties
#define BSDE_SOLIDTYPE		0x0140
#define BSDE_MOVEFLAGS		0x0141
#define BSDE_NOSPIN		0x0142	//solid can't rotate
#define BSDE_NOMOVE		0x0143	//solid can't be moved
#define BSDE_NOGRAVITY		0x0144	//solid is not effected by gravity
#define BSDE_STATIC		0x0145	//solid isn't effected by physics
#define BSDE_SEMISTATIC		0x0146	//solid may move but is
					//otherwise uneffected by physics
#define BSDE_FLUID		0x0147	//solid is a fluid
#define BSDE_NONSOLID		0x0148	//solid is non-solid
#define BSDE_TRIGGER		0x0149	//solid is a trigger

#define BSDE_NUM_FACES		0x014A	//number of faces (hull/trimesh)
#define BSDE_NUM_VERTICES	0x014B	//number of vertices (hull/trimesh)

//integer vector properties
//0x0160..0x017F
#define BSDE_FACE_VERTICES	0x0160	//face vertex numbers
#define BSDE_FACE_COUNTS	0x0161	//vertex counts for each face

//world properties

#define BSDE_GRAVITY		0x0180	//world gravity
#define BSDE_MAXSTEP		0x0181
#define BSDE_MINSTEP		0x0182
#define BSDE_MAXFRAME		0x0183
#define BSDE_REALTIME		0x0184
#define BSDE_WORLDTIME		0x0185

#define BSDE_CONTACT_SELF	0x0186	//contact self array
#define BSDE_CONTACT_OTHER	0x0187	//contact other array
#define BSDE_CONTACT_ORIGIN	0x0188	//contact origin array
#define BSDE_CONTACT_NORMAL	0x0189	//contact normal array
#define BSDE_TRIGGER_ARRAY	0x018A	//trigger array

#define BSDE_COLLIDE_STRICT	0x018B	//do not allow interpenetration
#define BSDE_FRAME_PASSES	0x018C

//solid types

#define BSDE_SOLID_OTHER	0x0210	//"other" solid type
#define BSDE_SOLID_AABB		0x0211
#define BSDE_SOLID_OBB		0x0212
#define BSDE_SOLID_SPHERE	0x0213
#define BSDE_SOLID_CYLINDER	0x0214	//flat topped cylinder
#define BSDE_SOLID_CCYLINDER	0x0215	//capped cylinder
#define BSDE_SOLID_CAPSULE	0x0215	//sphere capped cylinder
#define BSDE_SOLID_PLANE	0x0216	//normal and distance
#define BSDE_SOLID_TRIMESH	0x0217	//triangular mesh
#define BSDE_SOLID_HULL		0x0218	//convex hull
#define BSDE_SOLID_BSP		0x0219	//bsp tree


//joint types

#define BSDE_JOINT_OTHER	0x0240	//"other" joint type
#define BSDE_JOINT_BALL		0x0241	//keeps origins together
#define BSDE_JOINT_HINGE	0x0242	//works like a hinge
#define BSDE_JOINT_SLIDER	0x0243	//allows sliding along axis
#define BSDE_JOINT_UNIVERSAL	0x0244	//ball, keeps axis at 90 degrees
#define BSDE_JOINT_FIXED	0x0245	//fuses objects


//data types

#define BSDE_BYTE		0x0280
#define BSDE_UNSIGNED_BYTE	0x0281
#define BSDE_SHORT		0x0282
#define BSDE_UNSIGNED_SHORT	0x0283
#define BSDE_INT		0x0284
#define BSDE_UNSIGNED_INT	0x0285
#define BSDE_FLOAT		0x0286
#define BSDE_DOUBLE		0x0287

#define BSDE_2_BYTES		0x0288
#define BSDE_3_BYTES		0x0289
#define BSDE_4_BYTES		0x028A
#define BSDE_2_SHORTS		0x028B
#define BSDE_3_SHORTS		0x028C
#define BSDE_4_SHORTS		0x028D
#define BSDE_2_INTS		0x028E
#define BSDE_3_INTS		0x028F
#define BSDE_4_INTS		0x0290

#define BSDE_2_FLOATS		0x0291
#define BSDE_V2F		0x0291
#define BSDE_3_FLOATS		0x0292
#define BSDE_V3F		0x0292
#define BSDE_4_FLOATS		0x0293
#define BSDE_V4F		0x0293
#define BSDE_2_DOUBLES		0x0294
#define BSDE_V2D		0x0294
#define BSDE_3_DOUBLES		0x0295
#define BSDE_V3D		0x0295
#define BSDE_4_DOUBLES		0x0296
#define BSDE_V4D		0x0296


//construction related
#define	BSDE_BSP_LEAF		0x02C0
#define	BSDE_BSP_NODE		0x02C1
#define	BSDE_BSP_FACE		0x02C2

#endif
