/*
Thought:
Attempt to optimize iterator through the use of a quad tree.
Assume that, when the object counts get high enough that it should matter, the
quad tree will pay off.
*/

#include "bsde_internal.h"

//buffer for temporary nodes
BSDE_QuadNode *bsde_iter3_nodes=NULL;
int bsde_iter3_nodepos;

BSDE_BinaryNode *bsde_iter3_binodes=NULL;
int bsde_iter3_binodepos;

BSDE_QuadNode *BSDE_Iter3_AllocQuadNode()
{
	BSDE_QuadNode *tmp;
	if(!bsde_iter3_nodes)
	{
		bsde_iter3_nodes=malloc(4096*sizeof(BSDE_QuadNode));
		bsde_iter3_nodepos=0;
	}

	tmp=bsde_iter3_nodes+bsde_iter3_nodepos;
	bsde_iter3_nodepos++;
	return(tmp);
}

BSDE_BinaryNode *BSDE_Iter3_AllocBinaryNode()
{
	BSDE_BinaryNode *tmp;
	if(!bsde_iter3_binodes)
	{
		bsde_iter3_binodes=malloc(4096*sizeof(BSDE_BinaryNode));
		bsde_iter3_binodepos=0;
	}

	tmp=&bsde_iter3_binodes[bsde_iter3_binodepos++];
	return(tmp);
}

//calculate the centroid for the objects
int BSDE_Iter3_CalcCentroid(BSDE_Solid *first, double *org)
{
	BSDE_Solid *cur;
	int i;

	V3D_ZERO(org);
	cur=first;
	for(i=0; cur; i++)
	{
		V3D_ADD(org, cur->org, org);
		cur=cur->qnext;
	}
	V3D_SCALE(org, 1.0/(i?i:1), org);
	return(i);
}

//calculate the normal for the objects
int BSDE_Iter3_CalcNormal(BSDE_Solid *first, double *norm)
{
	double org[3], pt[3];
	BSDE_Solid *cur;
	int i;

	BSDE_Iter3_CalcCentroid(first, org);

	V3D_ZERO(norm);
	cur=first;
	for(i=0; cur; i++)
	{
		V3D_SUB(cur->org, org, pt);
		norm[0]+=fabs(pt[0]);
		norm[1]+=fabs(pt[1]);
		norm[2]+=fabs(pt[2]);
		cur=cur->qnext;
	}
	V3D_NORMALIZE(norm, norm);
	norm[3]=V3D_DOT(org, norm);

	return(i);
}

//collide an object with a list of objects
int BSDE_Iter3_CollideList(BSDE_World *world,
	BSDE_Solid *obj, BSDE_Solid *first, int n)
{
	BSDE_Solid *sa, *sb;
	int fla, flo, flx;
	int i, j, k, l;

	sa=obj;
	sb=first;
	while(sb)
	{
		if(sb->handle<=sa->handle)
		{
			sb=sb->qnext;
			continue;
		}

		fla=sa->moveflags&sb->moveflags;
		flo=sa->moveflags|sb->moveflags;
		flx=sa->moveflags^sb->moveflags;
		j=(fla&(BSDE_MOVEFL_STATIC|BSDE_MOVEFL_SEMISTATIC|
			BSDE_MOVEFL_NONSOLID)) &&
			!(flo&BSDE_MOVEFL_TRIGGER);
		if(j)
		{
			sb=sb->qnext;
			continue;
		}

		j=BSDE_Collide(world, sa, sb, world->contact+n);
		if(j>0)n+=j;

		if((j>0) && (sa->moveflags&BSDE_MOVEFL_TRIGGER))
			BSDE_WorldAddTriggerContact(world, sa, sb);
		if((j>0) && (sb->moveflags&BSDE_MOVEFL_TRIGGER))
			BSDE_WorldAddTriggerContact(world, sb, sa);
		sb=sb->qnext;
	}

	return(n);
}

//check collision with a list of objects
//m=0 for initial, 1 for final
int BSDE_Iter3_CheckCollideList(BSDE_World *world,
	BSDE_Solid *obj, BSDE_Solid *first, int m)
{
	BSDE_Solid *sa, *sb;
	int fla, flo, flx;
	int i, j, k, l;

	sa=obj;
	sb=first;
	while(sb)
	{
		fla=sa->moveflags&sb->moveflags;
		flo=sa->moveflags|sb->moveflags;
		flx=sa->moveflags^sb->moveflags;
		j=(fla&(BSDE_MOVEFL_STATIC|BSDE_MOVEFL_SEMISTATIC|
			BSDE_MOVEFL_NONSOLID)) &&
			!(flo&BSDE_MOVEFL_TRIGGER);
		if(j || (sa==sb))
		{
			sb=sb->qnext;
			continue;
		}

		j=BSDE_CheckCollide2(sa, sb, m);
		if(j)return(1);

		sb=sb->qnext;
	}

	return(0);
}


//quad trees

int BSDE_Iter3_SolidQuad(BSDE_Solid *solid, double *org)
{
	int i;

	i=0;
	if((solid->mmins[1]<org[1]) && (solid->mmins[0]<org[0]))i|=1;
	if((solid->mmins[1]<org[1]) && (solid->mmaxs[0]>org[0]))i|=2;
	if((solid->mmaxs[1]>org[1]) && (solid->mmins[0]<org[0]))i|=4;
	if((solid->mmaxs[1]>org[1]) && (solid->mmaxs[0]>org[0]))i|=8;

//	if((solid->mmins[0]<org[0]) && (solid->mmaxs[0]>org[0]))i|=16;
//	if((solid->mmins[1]<org[1]) && (solid->mmaxs[1]>org[1]))i|=16;

	return(i);
}

BSDE_QuadNode *BSDE_Iter3_BuildNode(BSDE_Solid *first)
{
	BSDE_QuadNode *tmp;
	BSDE_Solid *cur, *nxt, *l0, *l1, *l2, *l3, *lc;
	int i, j, k;

	tmp=BSDE_Iter3_AllocQuadNode();
	i=BSDE_Iter3_CalcCentroid(first, tmp->center);

	if(i<4)
	{
		tmp->node[0]=NULL;
		tmp->first=first;
		return(tmp);
	}

	l0=NULL; l1=NULL; l2=NULL; l3=NULL; lc=NULL;
	cur=first;
	while(cur)
	{
		nxt=cur->qnext;
		i=BSDE_Iter3_SolidQuad(cur, tmp->center);
		switch(i)
		{
		case 1:
			cur->qnext=l0;
			l0=cur;
			break;
		case 2:
			cur->qnext=l1;
			l1=cur;
			break;
		case 4:
			cur->qnext=l2;
			l2=cur;
			break;
		case 8:
			cur->qnext=l3;
			l3=cur;
			break;
		default:
			cur->qnext=lc;
			lc=cur;
			break;
		}
		cur=nxt;
	}

	if(l0 || l1 || l2 || l3)
	{
		tmp->node[0]=BSDE_Iter3_BuildNode(l0);
		tmp->node[1]=BSDE_Iter3_BuildNode(l1);
		tmp->node[2]=BSDE_Iter3_BuildNode(l2);
		tmp->node[3]=BSDE_Iter3_BuildNode(l3);
	}else
	{
		tmp->node[0]=NULL;
	}
	tmp->first=lc;

	return(tmp);
}

int BSDE_Iter3_CollideTree(BSDE_World *world, BSDE_QuadNode *node,
	BSDE_Solid *obj, int n)
{
	int i;

	if(!node->node[0])
	{
		n=BSDE_Iter3_CollideList(world, obj, node->first, n);
		return(n);
	}

	n=BSDE_Iter3_CollideList(world, obj, node->first, n);
	for(i=0; i<4; i++)
		if(BSDE_Iter3_SolidQuad(obj, node->center)&(1<<i))
			n=BSDE_Iter3_CollideTree(
				world, node->node[i], obj, n);
	return(n);
}


//BSP-tree like algo

BSDE_BinaryNode *BSDE_Iter3_BuildBiNode(BSDE_Solid *first)
{
	BSDE_BinaryNode *tmp;
	BSDE_Solid *cur, *nxt, *ll, *lr, *lc;
	double dm, dn;
	int i, j, k;

	tmp=BSDE_Iter3_AllocBinaryNode();
	tmp->up=NULL;	//in case root
	i=BSDE_Iter3_CalcNormal(first, tmp->norm);

	if(i<4)
	{
		cur=first;
		while(cur)
		{
			cur->node=tmp;
			cur=cur->qnext;
		}

		tmp->node[0]=NULL;
		tmp->first=first;

		return(tmp);
	}

	ll=NULL; lr=NULL; lc=NULL;
	cur=first;
	while(cur)
	{
		nxt=cur->qnext;
		BSDE_SolidAABB_BoxPlaneExtents(cur->mmins, cur->mmaxs,
			tmp->norm, &dm, &dn);
		if(dn<tmp->norm[3])
		{
			cur->qnext=ll;
			ll=cur;
		}else if(dm>tmp->norm[3])
		{
			cur->qnext=lr;
			lr=cur;
		}else
		{
			cur->node=tmp;
			cur->qnext=lc;
			lc=cur;
		}
		cur=nxt;
	}

	if(ll || lr)
	{
		tmp->node[0]=BSDE_Iter3_BuildBiNode(ll);
		tmp->node[1]=BSDE_Iter3_BuildBiNode(lr);
		tmp->node[0]->up=tmp;
		tmp->node[1]->up=tmp;

		tmp->first=lc;
	}else
	{
		tmp->node[0]=NULL;
		tmp->node[1]=NULL;
		tmp->first=lc;
	}

	return(tmp);
}

int BSDE_Iter3_CollideBiTree(BSDE_World *world, BSDE_BinaryNode *node,
	BSDE_Solid *obj, int n)
{
	double dm, dn;
	int i;

	if(!node->node[0])
	{
		n=BSDE_Iter3_CollideList(world, obj, node->first, n);
		return(n);
	}

	n=BSDE_Iter3_CollideList(world, obj, node->first, n);

	BSDE_SolidAABB_BoxPlaneExtents(obj->mmins, obj->mmaxs,
		node->norm, &dm, &dn);
	if(dn<node->norm[3])
	{
		n=BSDE_Iter3_CollideBiTree(world, node->node[0], obj, n);
	}else if(dm>node->norm[3])
	{
		n=BSDE_Iter3_CollideBiTree(world, node->node[1], obj, n);
	}else
	{
		n=BSDE_Iter3_CollideBiTree(world, node->node[0], obj, n);
		n=BSDE_Iter3_CollideBiTree(world, node->node[1], obj, n);
	}
	return(n);
}

int BSDE_Iter3_CollideBiObj(BSDE_World *world, BSDE_Solid *obj, int n)
{
	BSDE_BinaryNode *cur;
	n=BSDE_Iter3_CollideBiTree(world, obj->node, obj, n);

	cur=obj->node->up;
	while(cur)
	{
		n=BSDE_Iter3_CollideList(world, obj, cur->first, n);
		cur=cur->up;
	}
	return(n);
}

int BSDE_Iter3_CheckCollideBiTree(BSDE_World *world,
	BSDE_BinaryNode *node, BSDE_Solid *obj, int m)
{
	double dm, dn;
	int i;

	if(!node->node[0])
	{
		i=BSDE_Iter3_CheckCollideList(world, obj, node->first, m);
		return(i);
	}

	i=BSDE_Iter3_CheckCollideList(world, obj, node->first, m);
	if(i)return(i);

	BSDE_SolidAABB_BoxPlaneExtents(obj->mmins, obj->mmaxs,
		node->norm, &dm, &dn);
	if(dn<node->norm[3])
	{
		i=BSDE_Iter3_CheckCollideBiTree(world, node->node[0], obj, m);
		return(i);
	}else if(dm>node->norm[3])
	{
		i=BSDE_Iter3_CheckCollideBiTree(world, node->node[1], obj, m);
		return(i);
	}else
	{
		i=BSDE_Iter3_CheckCollideBiTree(world, node->node[0], obj, m);
		if(i)return(i);

		i=BSDE_Iter3_CheckCollideBiTree(world, node->node[1], obj, m);
		return(i);
	}
	return(0);
}

int BSDE_Iter3_CheckCollideBiObj(BSDE_World *world, BSDE_Solid *obj, int m)
{
	BSDE_BinaryNode *cur;
	int i;

	i=BSDE_Iter3_CheckCollideBiTree(world, obj->node, obj, m);
	if(i)return(i);

	cur=obj->node->up;
	while(cur)
	{
		i=BSDE_Iter3_CheckCollideList(world, obj, cur->first, m);
		if(i)return(i);
		cur=cur->up;
	}
	return(0);
}

int BSDE_WorldSingleStep2(BSDE_World *world, double dt)
{
	BSDE_Solid *sa, *sb;
	int fla, flo, flx;
	int i, j, k, l, n;

	sb=NULL;
	for(i=0; i<world->n_solids; i++)
	{
		sa=world->solids[i];
		if(!sa)continue;
		if(sa->moveflags&BSDE_MOVEFL_HIDDEN)continue;

		BSDE_Predict(world, sa, dt);

		sa->qnext=sb;
		sb=sa;
	}

	bsde_iter3_nodepos=0;
	bsde_iter3_binodepos=0;
//	world->root=BSDE_Iter3_BuildNode(sb);
	world->root=BSDE_Iter3_BuildBiNode(sb);

	j=world->passes;
	if(!j)j=8;
	while(j--)
	{
		n=0;
		for(i=0; i<world->n_solids; i++)
		{
			if(!world->solids[i])continue;
			sa=world->solids[i];
			if(sa->moveflags&BSDE_MOVEFL_HIDDEN)
				continue;

			l=BSDE_Iter3_CollideBiObj(world, sa, 0);
			n+=l;
			for(k=0; k<l; k++)
				BSDE_HandleContact(world,
					world->contact[k], dt);
			BSDE_FreeContacts();
		}
		if(!n)break;

		for(i=0; i<world->n_solids; i++)
		{
			sa=world->solids[i];
			if(!sa)continue;
			if(sa->moveflags&BSDE_MOVEFL_HIDDEN)continue;

			BSDE_UpdateVel(world, sa, dt);
			BSDE_Predict(world, sa, dt);
		}
	}

	j=world->passes;
	if(!j)j=1;
	while(j--)
	{
		for(i=0; i<world->n_joints; i++)
		{
			if(!world->joints[i])continue;
			BSDE_HandleJoint(world, world->joints[i], dt);

			sa=world->joints[i]->obja;
			sb=world->joints[i]->objb;

			BSDE_UpdateVel(world, sa, dt);
			BSDE_Predict(world, sa, dt);

			BSDE_UpdateVel(world, sb, dt);
			BSDE_Predict(world, sb, dt);
		}
	}

#if 1
	if(world->flags&BSDE_WORLDFL_STRICT)
		for(i=0; i<world->n_solids; i++)
	{
		if(!world->solids[i])continue;
		sa=world->solids[i];
		if(sa->moveflags&BSDE_MOVEFL_HIDDEN)
			continue;

		j=BSDE_Iter3_CheckCollideBiObj(world, sa, 1);
		if(j)
		{
			j=BSDE_Iter3_CheckCollideBiObj(world, sa, 0);
			if(j)
			{
//				if(sa->moveflags&BSDE_MOVEFL_BLOCKED)
//					printf("old initial collide\n");
//				printf("initial collide\n");

				j=16;
				while(j--)
				{
//					sa->org[0]+=BSDE_RandG()*0.1;
//					sa->org[1]+=BSDE_RandG()*0.1;
//					sa->org[2]+=BSDE_RandG()*0.1;
					sa->org[2]+=0.1;

					k=BSDE_Iter3_CheckCollideBiObj(
						world, sa, 0);
					if(!k)break;
				}

				//can't unstick, hide
				if(!j)
				{
					printf("can't unstick\n");
					sa->moveflags|=BSDE_MOVEFL_HIDDEN;
				}

				continue;
			}

			sa->moveflags|=BSDE_MOVEFL_BLOCKED;

			V4D_COPY(sa->org, sa->e_org);
			V4D_COPY(sa->rot, sa->e_rot);

			V3D_ZERO(sa->vel);
			V3D_ZERO(sa->tvel);
		}else sa->moveflags&=~BSDE_MOVEFL_BLOCKED;
	}
#endif

	for(i=0; i<world->n_solids; i++)
	{
		sa=world->solids[i];
		if(!sa)continue;
		if(sa->moveflags&BSDE_MOVEFL_HIDDEN)continue;

		BSDE_UpdateVel(world, sa, dt);
		BSDE_Update(world, sa, dt);
	}

	return(n);
}
