#include "bsde_internal.h"

BSDE_Joint *BSDE_NewJoint(int type)
{
	BSDE_Joint *tmp;

	tmp=malloc(sizeof(BSDE_Joint));
	memset(tmp, 0, sizeof(BSDE_Joint));
	tmp->type=type;

	return(tmp);
}

BSDE_Joint *BSDE_NextJointLink(BSDE_Joint *joint, BSDE_Solid *obj)
{
	if(joint->obja==obj)
	{
		return(joint->linka);
	}else if(joint->objb==obj)
	{
		return(joint->linkb);
	}

	printf("joint link error\n");
	return(NULL);
}

void BSDE_SetNextJointLink(BSDE_Joint *joint, BSDE_Solid *obj,
	BSDE_Joint *nxt)
{
	if(joint->obja==obj)
	{
		joint->linka=nxt;
	}else if(joint->objb==obj)
	{
		joint->linkb=nxt;
	}

	printf("joint link error\n");
}

void BSDE_UnlinkJointSolid(BSDE_Joint *joint, BSDE_Solid *obj)
{
	BSDE_Joint *cur, *nxt, *nxt2;

	cur=obj->joint;

	if(cur==joint)
	{
		nxt=BSDE_NextJointLink(cur, obj);
		obj->joint=nxt;

		BSDE_SetNextJointLink(cur, obj, NULL);
		return;
	}

	while(cur)
	{
		nxt=BSDE_NextJointLink(cur, obj);
		if(nxt==joint)
		{
			nxt2=BSDE_NextJointLink(cur, obj);
			BSDE_SetNextJointLink(cur, obj, nxt2);

			BSDE_SetNextJointLink(nxt, obj, NULL);
			return;
		}
	}

	printf("joint unlink error: not found\n");
	return;
}

void BSDE_BindJointSolids(BSDE_Joint *joint,
	BSDE_Solid *obja, BSDE_Solid *objb)
{
	if(joint->obja)BSDE_UnlinkJointSolid(joint, joint->obja);
	if(joint->objb)BSDE_UnlinkJointSolid(joint, joint->objb);

	joint->obja=obja;
	joint->objb=objb;

	joint->linka=obja->joint;
	joint->linkb=objb->joint;

	obja->joint=joint;
	objb->joint=joint;
}

void BSDE_JointConstrainPoints(BSDE_World *world,
	BSDE_Joint *joint, double *orga, double *orgb, double dt)
{
	double pa[3], pb[3], na[3];
	double fa[3], fb[3], va[3], vb[3];
	double to[3], tv[3];
	double r, f, g, tf;

	tf=50;

	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_NORMALIZE(va, na);

	V3D_ADDSCALE(orga, va, 0.5, to);

#if 0
	BSDE_GetPointVelocity(joint->obja, orga, pa);
	BSDE_GetPointVelocity(joint->objb, orgb, pb);

	V3D_COPY(joint->obja->vel, tv);
	V3D_ADD(tv, joint->objb->vel, tv);
	V3D_SCALE(tv, 0.5, tv);

	if(
		!(joint->obja->moveflags&
			(BSDE_MOVEFL_STATIC|BSDE_MOVEFL_SEMISTATIC))
		&&
		!(joint->objb->moveflags&
			(BSDE_MOVEFL_STATIC|BSDE_MOVEFL_SEMISTATIC))
		)
	{
		V3D_SUB(to, orga, fa);
		V3D_SUB(to, orgb, fb);
		V3D_ADD(joint->obja->org, fa, joint->obja->e_org);
		V3D_ADD(joint->objb->org, fb, joint->objb->e_org);

		V3D_COPY(tv, joint->obja->e_vel);
		V3D_COPY(tv, joint->objb->e_vel);
	}else if(!(joint->obja->moveflags&
		(BSDE_MOVEFL_STATIC|BSDE_MOVEFL_SEMISTATIC)))
	{
		V3D_ADD(joint->obja->org, va, joint->obja->e_org);
		V3D_COPY(joint->objb->vel, tv);
		V3D_COPY(tv, joint->obja->e_vel);
	}else
	{
		V3D_ADD(joint->objb->org, vb, joint->objb->e_org);
		V3D_COPY(joint->obja->vel, tv);
		V3D_COPY(tv, joint->objb->e_vel);
	}

	V3D_ZERO(joint->obja->tvel);
	V3D_ZERO(joint->objb->tvel);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);
	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);
#endif

#if 0
	BSDE_GetPointVelocity(joint->obja, orga, pa);
	BSDE_GetPointVelocity(joint->objb, orgb, pb);

	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
	r=f-g;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);

//	V3D_SCALE(na, -r, va);
//	V3D_SCALE(na, r, vb);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointForce(joint->obja, orga, fa);
	BSDE_ApplyPointForce(joint->objb, orgb, fb);
#endif

#if 1
	BSDE_GetVelocity(joint->obja, pa);
	BSDE_GetVelocity(joint->objb, pb);

	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
//	r=(f-g)*0.25;
	r=(f-g)*0.5;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);

	BSDE_ApplyAcceleration(joint->obja, va);
	BSDE_ApplyAcceleration(joint->objb, vb);


	BSDE_GetPointAngleVelocity(joint->obja, orga, pa);
	BSDE_GetPointAngleVelocity(joint->objb, orgb, pb);

	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
//	r=(f-g)*0.25;
	r=(f-g)*0.5;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);
#endif
}

void BSDE_JointConstrainPoints2(BSDE_World *world,
	BSDE_Joint *joint, double *orga, double *orgb, double dt)
{
	double pa[3], pb[3], na[3];
	double fa[3], fb[3], va[3], vb[3];
	double r, f, g;

	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_NORMALIZE(va, na);

	BSDE_GetPointAngleVelocity(joint->obja, orga, pa);
	BSDE_GetPointAngleVelocity(joint->objb, orgb, pb);

	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
	r=(f-g)*0.25;
//	r=(f-g)*0.5;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);
}

void BSDE_HandleJointFixed(BSDE_World *world, BSDE_Joint *joint, double dt)
{
	double orga[3], orgb[3];
	double org1a[3], org1b[3];
	double org2a[3], org2b[3];
	double org3a[3], org3b[3];

	if(!joint->obja)return;
	if(!joint->objb)return;

	BSDE_PlaneD_TransformPoint(joint->orga, joint->obja->e_xform, orga);
	BSDE_PlaneD_TransformPoint(joint->orgb, joint->objb->e_xform, orgb);
	BSDE_PlaneD_TransformPoint(joint->org1a, joint->obja->e_xform, org1a);
	BSDE_PlaneD_TransformPoint(joint->org1b, joint->objb->e_xform, org1b);
	BSDE_PlaneD_TransformPoint(joint->org2a, joint->obja->e_xform, org2a);
	BSDE_PlaneD_TransformPoint(joint->org2b, joint->objb->e_xform, org2b);
	BSDE_PlaneD_TransformPoint(joint->org3a, joint->obja->e_xform, org3a);
	BSDE_PlaneD_TransformPoint(joint->org3b, joint->objb->e_xform, org3b);

	BSDE_JointConstrainPoints(world, joint, orga, orgb, dt);
	BSDE_JointConstrainPoints2(world, joint, org1a, org1b, dt);
	BSDE_JointConstrainPoints2(world, joint, org2a, org2b, dt);
	BSDE_JointConstrainPoints2(world, joint, org3a, org3b, dt);
}

void BSDE_HandleJointBall(BSDE_World *world, BSDE_Joint *joint, double dt)
{
	double orga[3], orgb[3], org[3];
	double pa[3], pb[3], na[3], nb[3];
	double fa[3], fb[3], va[3], vb[3];
	double r, f, g;

	if(!joint->obja)return;
	if(!joint->objb)return;
//	printf("handle joint ball\n");

	BSDE_PlaneD_TransformPoint(
		joint->orga, joint->obja->e_xform, orga);
	BSDE_PlaneD_TransformPoint(
		joint->orgb, joint->objb->e_xform, orgb);

//	printf("orga %f %f %f\n", orga[0], orga[1], orga[2]);
//	printf("orgb %f %f %f\n", orgb[0], orgb[1], orgb[2]);

	BSDE_GetVelocity(joint->obja, pa);
	BSDE_GetVelocity(joint->objb, pb);

	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);

//	V3D_SCALE(va, 0.5, va);
//	V3D_SCALE(vb, 0.5, vb);


	V3D_NORMALIZE(va, na);
//	V3D_NORMALIZE(vb, nb);
	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
	r=(f-g)*0.25;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);

//	printf("fa %f %f %f\n", fa[0], fa[1], fa[2]);
//	printf("fb %f %f %f\n", fb[0], fb[1], fb[2]);

	BSDE_ApplyAcceleration(joint->obja, va);
	BSDE_ApplyAcceleration(joint->objb, vb);


	BSDE_GetPointAngleVelocity(joint->obja, orga, pa);
	BSDE_GetPointAngleVelocity(joint->objb, orgb, pb);

	f=V3D_DOT(pa, na);
	g=V3D_DOT(pb, na);
	r=(f-g)*0.25;
	V3D_ADDSCALE(va, na, -r, va);
	V3D_ADDSCALE(vb, na, r, vb);


	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);
}

void BSDE_HandleJointHinge(BSDE_World *world, BSDE_Joint *joint, double dt)
{
	double orga[3], orgb[3], org[3];
	double dira[3], dirb[3], dir[3];
	double org2a[3], org2b[3];
	double lpa[3], lpb[3];
	double apa[3], apb[3];
	double ap2a[3], ap2b[3];
	double fa[3], fb[3], va[3], vb[3];
	double r;

	if(!joint->obja)return;
	if(!joint->objb)return;
//	printf("handle joint hinge\n");

	BSDE_PlaneD_TransformPoint(
		joint->orga, joint->obja->e_xform, orga);
	BSDE_PlaneD_TransformPoint(
		joint->orgb, joint->objb->e_xform, orgb);
	BSDE_PlaneD_TransformVector(
		joint->dira, joint->obja->e_xform, dira);
	BSDE_PlaneD_TransformVector(
		joint->dirb, joint->objb->e_xform, dirb);

	V3D_ADDSCALE(orga, dira, 100, org2a);
	V3D_ADDSCALE(orgb, dirb, 100, org2b);

	BSDE_GetVelocity(joint->obja, lpa);
	BSDE_GetVelocity(joint->objb, lpb);
	BSDE_GetPointAngleVelocity(joint->obja, orga, apa);
	BSDE_GetPointAngleVelocity(joint->objb, orgb, apb);
	BSDE_GetPointAngleVelocity(joint->obja, org2a, ap2a);
	BSDE_GetPointAngleVelocity(joint->objb, org2b, ap2b);

	//linear force at orgs
	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_SCALE(va, 1, va);
	V3D_SCALE(vb, 1, vb);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyForce(joint->obja, fa);
	BSDE_ApplyForce(joint->objb, fb);

	//angular force at orgs
	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);

	//angular force at org2
	V3D_SUB(org2b, org2a, va);
	V3D_SUB(org2a, org2b, vb);
//	V3D_SUB(org2a, org2b, va);
//	V3D_SUB(org2b, org2a, vb);
	V3D_SCALE(va, joint->obja->mass*dt, fa);
	V3D_SCALE(vb, joint->objb->mass*dt, fb);

	BSDE_ApplyPointAngleForce(joint->obja, org2a, fa);
	BSDE_ApplyPointAngleForce(joint->objb, org2b, fb);
}

void BSDE_HandleJointUniversal(BSDE_World *world,
	BSDE_Joint *joint, double dt)
{
	double orga[3], orgb[3], org[3];
	double dira[3], dirb[3], dir[3];
	double dir2a[3], dir2b[3];
	double org2a[3], org2b[3];
	double org3a[3], org3b[3];
	double lpa[3], lpb[3];
	double apa[3], apb[3];
	double ap2a[3], ap2b[3];
	double fa[3], fb[3], va[3], vb[3];
	double r;

	if(!joint->obja)return;
	if(!joint->objb)return;
//	printf("handle joint universal\n");

	BSDE_PlaneD_TransformPoint(
		joint->orga, joint->obja->e_xform, orga);
	BSDE_PlaneD_TransformPoint(
		joint->orgb, joint->objb->e_xform, orgb);
	BSDE_PlaneD_TransformVector(
		joint->dira, joint->obja->e_xform, dira);
	BSDE_PlaneD_TransformVector(
		joint->dirb, joint->objb->e_xform, dirb);

	V3D_CROSS(dira, dirb, dir);
	V3D_CROSS(dir, dira, dir2b);
	V3D_CROSS(dirb, dir, dir2a);

	V3D_NORMALIZE(dir2a, dir2a);
	V3D_NORMALIZE(dir2b, dir2b);

//	printf("A (%f %f %f) (%f %f %f)\n", dira[0], dira[1], dira[2],
//		dir2a[0], dir2a[1], dir2a[2]);
//	printf("B (%f %f %f) (%f %f %f)\n", dirb[0], dirb[1], dirb[2],
//		dir2b[0], dir2b[1], dir2b[2]);

	V3D_ADDSCALE(orga, dira, 100, org2a);
	V3D_ADDSCALE(orgb, dirb, 100, org2b);

	V3D_ADDSCALE(orga, dir2a, 100, org3a);
	V3D_ADDSCALE(orgb, dir2b, 100, org3b);


	BSDE_GetVelocity(joint->obja, lpa);
	BSDE_GetVelocity(joint->objb, lpb);
	BSDE_GetPointAngleVelocity(joint->obja, orga, apa);
	BSDE_GetPointAngleVelocity(joint->objb, orgb, apb);
	BSDE_GetPointAngleVelocity(joint->obja, org2a, ap2a);
	BSDE_GetPointAngleVelocity(joint->objb, org2b, ap2b);

	//linear force at orgs
	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_SCALE(va, 1, va);
	V3D_SCALE(vb, 1, vb);

	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyForce(joint->obja, fa);
	BSDE_ApplyForce(joint->objb, fb);

	//angular force at orgs
	V3D_SUB(orgb, orga, va);
	V3D_SUB(orga, orgb, vb);
	V3D_SCALE(va, joint->obja->mass, fa);
	V3D_SCALE(vb, joint->objb->mass, fb);

	BSDE_ApplyPointAngleForce(joint->obja, orga, fa);
	BSDE_ApplyPointAngleForce(joint->objb, orgb, fb);

	//angular force at org2
	V3D_SUB(org3a, org2a, va);
	V3D_SUB(org3b, org2b, vb);
	V3D_SCALE(va, joint->obja->mass*dt, fa);
	V3D_SCALE(vb, joint->objb->mass*dt, fb);

	BSDE_ApplyPointAngleForce(joint->obja, org2a, fa);
	BSDE_ApplyPointAngleForce(joint->objb, org2b, fb);
}

void BSDE_HandleJoint(BSDE_World *world, BSDE_Joint *joint, double dt)
{
//	printf("handle joint\n");

	switch(joint->type)
	{
	case BSDE_JOINT_BALL:
		BSDE_HandleJointBall(world, joint, dt);
		break;
	case BSDE_JOINT_HINGE:
		BSDE_HandleJointHinge(world, joint, dt);
		break;
	case BSDE_JOINT_UNIVERSAL:
		BSDE_HandleJointUniversal(world, joint, dt);
		break;
	case BSDE_JOINT_FIXED:
		BSDE_HandleJointFixed(world, joint, dt);
		break;
	default:
		break;
	}

	if(joint->obja)BSDE_UpdateVel(world, joint->obja, dt);
	if(joint->objb)BSDE_UpdateVel(world, joint->objb, dt);
}

double BSDE_GetJointAttrF(BSDE_Joint *self, int var)
{
	double f;

	f=0;
	switch(var)
	{
	default:
		break;
	}
	return(f);
}

void BSDE_SetJointAttrF(BSDE_Joint *self, int var, double val)
{
	switch(var)
	{
	default:
		break;
	}
}

int BSDE_GetJointAttrI(BSDE_Joint *self, int var)
{
	int i;

	i=0;
	switch(var)
	{
	default:
		break;
	}
	return(i);
}

void BSDE_SetJointAttrI(BSDE_Joint *self, int var, int val)
{
	int i;
	switch(var)
	{
	default:
		break;
	}
}

void BSDE_GetJointAttrFV(BSDE_Joint *self, int var, float *val, int n)
{
	double vec[64];
	int i;

	BSDE_GetJointAttrDV(self, var, vec, n);
	for(i=0; i<n; i++)val[i]=vec[i];
}

void BSDE_SetJointAttrFV(BSDE_Joint *self, int var, float *val, int n)
{
	double vec[64];
	int i;

	for(i=0; i<n; i++)vec[i]=val[i];
	BSDE_SetJointAttrDV(self, var, vec, n);
}

void BSDE_GetJointAttrDV(BSDE_Joint *self, int var, double *val, int n)
{
	switch(var)
	{
	default:
		break;
	}
}

void BSDE_SetJointAttrDV(BSDE_Joint *self, int var, double *val, int n)
{
	double org[3];

	switch(var)
	{
	case BSDE_ORIGIN:
		BSDE_TransformPointLocal(self->obja, val, self->orga);
		BSDE_TransformPointLocal(self->objb, val, self->orgb);
		break;
	case BSDE_ORIGIN_1:
		BSDE_TransformPointLocal(self->obja, val, self->org1a);
		BSDE_TransformPointLocal(self->objb, val, self->org1b);
		break;
	case BSDE_ORIGIN_2:
		BSDE_TransformPointLocal(self->obja, val, self->org2a);
		BSDE_TransformPointLocal(self->objb, val, self->org2b);
		break;
	case BSDE_ORIGIN_3:
		BSDE_TransformPointLocal(self->obja, val, self->org3a);
		BSDE_TransformPointLocal(self->objb, val, self->org3b);
		break;
	case BSDE_ORIGIN_123:
		BSDE_TransformPointLocal(self->obja, val, self->orga);
		BSDE_TransformPointLocal(self->objb, val, self->orgb);
		V3D_COPY(val, org); org[0]+=1;
		BSDE_TransformPointLocal(self->obja, org, self->org1a);
		BSDE_TransformPointLocal(self->objb, org, self->org1b);
		V3D_COPY(val, org); org[1]+=1;
		BSDE_TransformPointLocal(self->obja, org, self->org2a);
		BSDE_TransformPointLocal(self->objb, org, self->org2b);
		V3D_COPY(val, org); org[2]+=1;
		BSDE_TransformPointLocal(self->obja, org, self->org3a);
		BSDE_TransformPointLocal(self->objb, org, self->org3b);
		break;

	case BSDE_ORIGIN_A:
		BSDE_TransformPointLocal(self->obja, val, self->orga);
		break;
	case BSDE_ORIGIN_B:
		BSDE_TransformPointLocal(self->objb, val, self->orgb);
		break;
	case BSDE_ORIGIN_RA:
		V3D_COPY(val, self->orga);
		break;
	case BSDE_ORIGIN_RB:
		V3D_COPY(val, self->orgb);
		break;

	case BSDE_DIRECTION:
		BSDE_TransformVectorLocal(self->obja, val, self->dira);
		BSDE_TransformVectorLocal(self->objb, val, self->dirb);
		break;
	case BSDE_DIRECTION_1:
		BSDE_TransformVectorLocal(self->obja, val, self->dir1a);
		BSDE_TransformVectorLocal(self->objb, val, self->dir1b);
		break;

	case BSDE_DIRECTION_A:
		BSDE_TransformVectorLocal(self->obja, val, self->dira);
		break;
	case BSDE_DIRECTION_B:
		BSDE_TransformVectorLocal(self->objb, val, self->dirb);
		break;
	case BSDE_DIRECTION_RA:
		V3D_COPY(val, self->orga);
		break;
	case BSDE_DIRECTION_RB:
		V3D_COPY(val, self->orgb);
		break;

	default:
		break;
	}
}
