#ifndef BSDE_MATH_H
#define BSDE_MATH_H

#ifndef INLINE
#define INLINE
#endif

#define BSDE_M_PI		3.1415926535897932384626433832795
#define BSDE_M_BIGVAL		1000000000.0	//some large value
#define BSDE_M_SMALLVAL		0.0000000001	//some small value

#define DEG(x)		((x)*(BSDE_M_PI/180.0))
#define MIN(x, y)	(((x)<(y))?(x):(y))
#define MAX(x, y)	(((x)>(y))?(x):(y))


//macro varieties
#define V2F_COPY(a, b)		(b)[0]=(a)[0];	\
				(b)[1]=(a)[1];
#define V3F_COPY(a, b)		(b)[0]=(a)[0];	\
				(b)[1]=(a)[1];	\
				(b)[2]=(a)[2];
#define V4F_COPY(a, b)		(b)[0]=(a)[0];	\
				(b)[1]=(a)[1];	\
				(b)[2]=(a)[2];	\
				(b)[3]=(a)[3];
#define V2D_COPY(a, b)		V2F_COPY(a, b)
#define V3D_COPY(a, b)		V3F_COPY(a, b)
#define V4D_COPY(a, b)		V4F_COPY(a, b)
#define V2D_ADD(a, b, c)	V2F_ADD(a, b, c)
#define V3D_ADD(a, b, c)	V3F_ADD(a, b, c)
#define V4D_ADD(a, b, c)	V4F_ADD(a, b, c)

#define V2F_ADD(a, b, c)	(c)[0]=(a)[0]+(b)[0];	\
				(c)[1]=(a)[1]+(b)[1];
#define V3F_ADD(a, b, c)	(c)[0]=(a)[0]+(b)[0];	\
				(c)[1]=(a)[1]+(b)[1];	\
				(c)[2]=(a)[2]+(b)[2];
#define V4F_ADD(a, b, c)	(c)[0]=(a)[0]+(b)[0];	\
				(c)[1]=(a)[1]+(b)[1];	\
				(c)[2]=(a)[2]+(b)[2];	\
				(c)[3]=(a)[3]+(b)[3];

#define V2F_SUB(a, b, c)	(c)[0]=(a)[0]-(b)[0];	\
				(c)[1]=(a)[1]-(b)[1];
#define V3F_SUB(a, b, c)	(c)[0]=(a)[0]-(b)[0];	\
				(c)[1]=(a)[1]-(b)[1];	\
				(c)[2]=(a)[2]-(b)[2];
#define V4F_SUB(a, b, c)	(c)[0]=(a)[0]-(b)[0];	\
				(c)[1]=(a)[1]-(b)[1];	\
				(c)[2]=(a)[2]-(b)[2];	\
				(c)[3]=(a)[3]-(b)[3];

#define V2F_SCALE(a, b, c)	(c)[0]=(a)[0]*(b);	\
				(c)[1]=(a)[1]*(b);
#define V3F_SCALE(a, b, c)	(c)[0]=(a)[0]*(b);	\
				(c)[1]=(a)[1]*(b);	\
				(c)[2]=(a)[2]*(b);
#define V4F_SCALE(a, b, c)	(c)[0]=(a)[0]*(b);	\
				(c)[1]=(a)[1]*(b);	\
				(c)[2]=(a)[2]*(b);	\
				(c)[3]=(a)[3]*(b);

#define V2F_SCALEVEC(a, b, c)	(c)[0]=(a)[0]*(b)[0];	\
				(c)[1]=(a)[1]*(b)[1];
#define V3F_SCALEVEC(a, b, c)	(c)[0]=(a)[0]*(b)[0];	\
				(c)[1]=(a)[1]*(b)[1];	\
				(c)[2]=(a)[2]*(b)[2];
#define V4F_SCALEVEC(a, b, c)	(c)[0]=(a)[0]*(b)[0];	\
				(c)[1]=(a)[1]*(b)[1];	\
				(c)[2]=(a)[2]*(b)[2];	\
				(c)[3]=(a)[3]*(b)[3];

#define V2F_DOT(a, b)		((a)[0]*(b)[0]+(a)[1]*(b)[1])
#define V3F_DOT(a, b)		((a)[0]*(b)[0]+(a)[1]*(b)[1]+(a)[2]*(b)[2])
#define V4F_DOT(a, b)		((a)[0]*(b)[0]+(a)[1]*(b)[1]+ \
				(a)[2]*(b)[2]+(a)[3]*(b)[3])

#define V2F_CROSS(a, b)		((a)[0]*(b)[1]-(a)[1]*(b)[0])
#define V3F_CROSS(a, b, c)	(c)[0]=((a)[1]*(b)[2])-((a)[2]*(b)[1]);	\
				(c)[1]=((a)[2]*(b)[0])-((a)[0]*(b)[2]);	\
				(c)[2]=((a)[0]*(b)[1])-((a)[1]*(b)[0]);

#define V2F_LEN(a)		(sqrt(V2F_DOT(a, a)))
#define V3F_LEN(a)		(sqrt(V3F_DOT(a, a)))
#define V4F_LEN(a)		(sqrt(V4F_DOT(a, a)))

#define V1F_DIST2(a, b)		(((a)-(b))*((a)-(b)))
#define V2F_DIST2(a, b)		(V1F_DIST2((a)[0], (b)[0])+ \
				V1F_DIST2((a)[1], (b)[1]))
#define V3F_DIST2(a, b)		(V1F_DIST2((a)[0], (b)[0])+ \
				V1F_DIST2((a)[1], (b)[1])+ \
				V1F_DIST2((a)[2], (b)[2]))
#define V4F_DIST2(a, b)		(V1F_DIST2((a)[0], (b)[0])+ \
				V1F_DIST2((a)[1], (b)[1])+ \
				V1F_DIST2((a)[2], (b)[2])+ \
				V1F_DIST2((a)[3], (b)[3]))

#define V2F_DIST(a, b)		(sqrt(V2F_DIST2(a, b)))
#define V3F_DIST(a, b)		(sqrt(V3F_DIST2(a, b)))
#define V4F_DIST(a, b)		(sqrt(V4F_DIST2(a, b)))


#define V2F_ZERO(a)		(a)[0]=0; (a)[1]=0;


#define V3F_ZERO(a)		(a)[0]=0; (a)[1]=0; (a)[2]=0;


#define V4F_ZERO(a)		(a)[0]=0; (a)[1]=0; (a)[2]=0; (a)[3]=0;



#define V2F_CONST(a, k)		(a)[0]=k; (a)[1]=k;


#define V3F_CONST(a, k)		(a)[0]=k; (a)[1]=k; (a)[2]=k;


#define V4F_CONST(a, k)		(a)[0]=k; (a)[1]=k; (a)[2]=k; (a)[3]=k;

#define V2F_SET(a, x, y)	(a)[0]=x; (a)[1]=y;
#define V3F_SET(a, x, y, z)	(a)[0]=x; (a)[1]=y; (a)[2]=z;
#define V4F_SET(a, x, y, z, w)	(a)[0]=x; (a)[1]=y; (a)[2]=z; (a)[3]=w;

#define V2D_SET(a, x, y)	(a)[0]=x; (a)[1]=y;
#define V3D_SET(a, x, y, z)	(a)[0]=x; (a)[1]=y; (a)[2]=z;
#define V4D_SET(a, x, y, z, w)	(a)[0]=x; (a)[1]=y; (a)[2]=z; (a)[3]=w;

#define V2F_ADDCONST(a, k, b)	(b)[0]=(a)[0]+(k); (b)[1]=(a)[1]+(k);


#define V3F_ADDCONST(a, k, b)	(b)[0]=(a)[0]+(k); (b)[1]=(a)[1]+(k); \
				(b)[2]=(a)[2]+(k);


#define V4F_ADDCONST(a, k, b)	(b)[0]=(a)[0]+(k); (b)[1]=(a)[1]+(k); \
				(b)[2]=(a)[2]+(k); (b)[3]=(a)[3]+(k);


#define V2F_SUBCONST(a, k, b)	(b)[0]=(a)[0]-(k); (b)[1]=(a)[1]-(k);


#define V3F_SUBCONST(a, k, b)	(b)[0]=(a)[0]-(k); (b)[1]=(a)[1]-(k); \
				(b)[2]=(a)[2]-(k);


#define V4F_SUBCONST(a, k, b)	(b)[0]=(a)[0]-(k); (b)[1]=(a)[1]-(k); \
				(b)[2]=(a)[2]-(k); (b)[3]=(a)[3]-(k);


#define V2F_ADDSCALE(a, b, c, d)	(d)[0]=(a)[0]+(b)[0]*(c);	\
					(d)[1]=(a)[1]+(b)[1]*(c);
#define V3F_ADDSCALE(a, b, c, d)	(d)[0]=(a)[0]+(b)[0]*(c);	\
					(d)[1]=(a)[1]+(b)[1]*(c);	\
					(d)[2]=(a)[2]+(b)[2]*(c);
#define V4F_ADDSCALE(a, b, c, d)	(d)[0]=(a)[0]+(b)[0]*(c);	\
					(d)[1]=(a)[1]+(b)[1]*(c);	\
					(d)[2]=(a)[2]+(b)[2]*(c);	\
					(d)[3]=(a)[3]+(b)[3]*(c);

#define V2F_SCALEADDSCALE(a, b, c, d, e)	\
					(e)[0]=(a)[0]*(b)+(c)[0]*(d);	\
					(e)[1]=(a)[1]*(b)+(c)[1]*(d);
#define V3F_SCALEADDSCALE(a, b, c, d, e)	\
					(e)[0]=(a)[0]*(b)+(c)[0]*(d);	\
					(e)[1]=(a)[1]*(b)+(c)[1]*(d);	\
					(e)[2]=(a)[2]*(b)+(c)[2]*(d);
#define V4F_SCALEADDSCALE(a, b, c, d, e)	\
					(e)[0]=(a)[0]*(b)+(c)[0]*(d);	\
					(e)[1]=(a)[1]*(b)+(c)[1]*(d);	\
					(e)[2]=(a)[2]*(b)+(c)[2]*(d);	\
					(e)[3]=(a)[3]*(B)+(c)[3]*(d);

#include "bsde_math_dot.h"

#define V2D_COPY(a, b)		V2F_COPY(a, b)
#define V3D_COPY(a, b)		V3F_COPY(a, b)
#define V4D_COPY(a, b)		V4F_COPY(a, b)
#define V2D_ADD(a, b, c)	V2F_ADD(a, b, c)
#define V3D_ADD(a, b, c)	V3F_ADD(a, b, c)
#define V4D_ADD(a, b, c)	V4F_ADD(a, b, c)
#define V2D_SUB(a, b, c)	V2F_SUB(a, b, c)
#define V3D_SUB(a, b, c)	V3F_SUB(a, b, c)
#define V4D_SUB(a, b, c)	V4F_SUB(a, b, c)
#define V2D_SCALE(a, b, c)	V2F_SCALE(a, b, c)
#define V3D_SCALE(a, b, c)	V3F_SCALE(a, b, c)
#define V4D_SCALE(a, b, c)	V4F_SCALE(a, b, c)
#define V2D_SCALEVEC(a, b, c)	V2F_SCALEVEC(a, b, c)
#define V3D_SCALEVEC(a, b, c)	V3F_SCALEVEC(a, b, c)
#define V4D_SCALEVEC(a, b, c)	V4F_SCALEVEC(a, b, c)
#define V2D_DOT(a, b)		V2F_DOT(a, b)
#define V3D_DOT(a, b)		V3F_DOT(a, b)
#define V4D_DOT(a, b)		V4F_DOT(a, b)
#define V2D_CROSS(a, b)		V2F_CROSS(a, b)
#define V3D_CROSS(a, b, c)	V3F_CROSS(a, b, c)
#define V2D_LEN(a)		V2F_LEN(a)
#define V3D_LEN(a)		V3F_LEN(a)
#define V4D_LEN(a)		V4F_LEN(a)
#define V2D_DIST(a, b)		V2F_DIST(a, b)
#define V3D_DIST(a, b)		V3F_DIST(a, b)
#define V4D_DIST(a, b)		V4F_DIST(a, b)
#define V2D_ZERO(a)		V2F_ZERO(a)
#define V3D_ZERO(a)		V3F_ZERO(a)
#define V4D_ZERO(a)		V4F_ZERO(a)
#define V2D_CONST(a, k)		V2F_CONST(a, k)
#define V3D_CONST(a, k)		V3F_CONST(a, k)
#define V4D_CONST(a, k)		V4F_CONST(a, k)
#define V2D_ADDCONST(a, k, b)	V2F_ADDCONST(a, k, b)
#define V3D_ADDCONST(a, k, b)	V3F_ADDCONST(a, k, b)
#define V4D_ADDCONST(a, k, b)	V4F_ADDCONST(a, k, b)
#define V2D_SUBCONST(a, k, b)	V2F_SUBCONST(a, k, b)
#define V3D_SUBCONST(a, k, b)	V3F_SUBCONST(a, k, b)
#define V4D_SUBCONST(a, k, b)	V4F_SUBCONST(a, k, b)
#define V2D_ADDSCALE(a, b, c, d)	V2F_ADDSCALE(a, b, c, d)
#define V3D_ADDSCALE(a, b, c, d)	V3F_ADDSCALE(a, b, c, d)
#define V4D_ADDSCALE(a, b, c, d)	V4F_ADDSCALE(a, b, c, d)

#define V2D_SCALEADDSCALE(a, b, c, d, e)	\
	V2F_SCALEADDSCALE(a, b, c, d, e)
#define V3D_SCALEADDSCALE(a, b, c, d, e)	\
	V3F_SCALEADDSCALE(a, b, c, d, e)
#define V4D_SCALEADDSCALE(a, b, c, d, e)	\
	V4F_SCALEADDSCALE(a, b, c, d, e)

#define V2F_SUM(a)		((a)[0]+(a)[1])
#define V3F_SUM(a)		((a)[0]+(a)[1]+(a)[2])
#define V4F_SUM(a)		((a)[0]+(a)[1]+(a)[2]+(a)[3])

#define V2D_INITMINS(a)		V2D_CONST(a, BSDE_M_BIGVAL)
#define V2D_INITMAXS(a)		V2D_CONST(a, -BSDE_M_BIGVAL)
#define V3D_INITMINS(a)		V3D_CONST(a, BSDE_M_BIGVAL)
#define V3D_INITMAXS(a)		V3D_CONST(a, -BSDE_M_BIGVAL)
#define V4D_INITMINS(a)		V4D_CONST(a, BSDE_M_BIGVAL)
#define V4D_INITMAXS(a)		V4D_CONST(a, -BSDE_M_BIGVAL)

#define V2D_INITBOX(m, n)	V2D_INITMINS(m) V2D_INITMAXS(n)
#define V3D_INITBOX(m, n)	V3D_INITMINS(m) V3D_INITMAXS(n)
#define V4D_INITBOX(m, n)	V4D_INITMINS(m) V4D_INITMAXS(n)

#define V2D_MIN(a, b, c)	(c)[0]=MIN((a)[0], (b)[0]); \
				(c)[1]=MIN((a)[1], (b)[1]);
#define V3D_MIN(a, b, c)	(c)[0]=MIN((a)[0], (b)[0]); \
				(c)[1]=MIN((a)[1], (b)[1]); \
				(c)[2]=MIN((a)[2], (b)[2]);
#define V4D_MIN(a, b, c)	(c)[0]=MIN((a)[0], (b)[0]); \
				(c)[1]=MIN((a)[1], (b)[1]); \
				(c)[2]=MIN((a)[2], (b)[2]); \
				(c)[3]=MIN((a)[3], (b)[3]);
#define V2D_MAX(a, b, c)	(c)[0]=MAX((a)[0], (b)[0]); \
				(c)[1]=MAX((a)[1], (b)[1]);
#define V3D_MAX(a, b, c)	(c)[0]=MAX((a)[0], (b)[0]); \
				(c)[1]=MAX((a)[1], (b)[1]); \
				(c)[2]=MAX((a)[2], (b)[2]);
#define V4D_MAX(a, b, c)	(c)[0]=MAX((a)[0], (b)[0]); \
				(c)[1]=MAX((a)[1], (b)[1]); \
				(c)[2]=MAX((a)[2], (b)[2]); \
				(c)[3]=MAX((a)[3], (b)[3]);

static INLINE float V2F_NORMALIZE(float *a, float *b)
	{float f, g; f=V2F_LEN(a); g=(f==0)?1:(1.0/f);
	V2F_SCALE(a, g, b); return(f);}

static INLINE float V3F_NORMALIZE(float *a, float *b)
	{float f, g; f=V3F_LEN(a); g=(f==0)?1:(1.0/f);
	V3F_SCALE(a, g, b); return(f);}

static INLINE float V4F_NORMALIZE(float *a, float *b)
	{float f, g; f=V4F_LEN(a); g=(f==0)?1:(1.0/f);
	V4F_SCALE(a, g, b); return(f);}

static INLINE double V2D_NORMALIZE(double *a, double *b)
	{double f, g; f=V2F_LEN(a); g=(f==0)?1:(1.0/f);
	V2F_SCALE(a, g, b); return(f);}
static INLINE double V3D_NORMALIZE(double *a, double *b)
	{double f, g; f=V3F_LEN(a); g=(f==0)?1:(1.0/f);
	V3F_SCALE(a, g, b); return(f);}
static INLINE double V4D_NORMALIZE(double *a, double *b)
	{double f, g; f=V4F_LEN(a); g=(f==0)?1:(1.0/f);
	V4F_SCALE(a, g, b); return(f);}

#endif
