#include "bsde_internal.h"

void BSDE_StepWorld(BSDE_World *world, double dt)
{
	int i;

	world->realtime+=dt;

	dt+=world->steptime;
	if(dt>world->maxframe)
	{
		world->steptime=dt-world->maxframe;
		dt=world->maxframe;
	}

	//forget about backlog
	i=0;
	while(world->steptime>world->maxframe)
	{
		world->steptime-=world->maxframe;
		i++;
	}
//	if(i)printf("dropped %d frames\n", i);

//	dt+=world->steptime;
//	if(dt>world->maxframe)dt=world->maxframe;

	while(dt>=world->maxstep)
	{
		BSDE_WorldForces(world, world->maxstep);
//		BSDE_SetupStack(world);
//		BSDE_PushInitial(world);
//		BSDE_RecurseWorld(world, world->maxstep, 0);

//		BSDE_WorldMultiStep(world, world->maxstep);
		BSDE_WorldSingleStep2(world, world->maxstep);

		dt-=world->maxstep;
		world->worldtime+=world->maxstep;
	}

	//accumulate remaining partial step
	world->steptime=dt;
}

void BSDE_WorldForces(BSDE_World *world, double dt)
{
	double dv[3];
	BSDE_Solid *cur;
	int i;

	for(i=0; i<world->n_solids; i++)
	{
		if(!world->solids[i])continue;
		cur=world->solids[i];
		if(cur->moveflags&BSDE_MOVEFL_HIDDEN)continue;

		if(cur->moveflags&BSDE_MOVEFL_BLOCKED)
		{
//			cur->vel[0]+=BSDE_RandG()*0.1;
//			cur->vel[1]+=BSDE_RandG()*0.1;
//			cur->vel[2]+=BSDE_RandG()*0.1;
			continue;
		}

		if(!(cur->moveflags&BSDE_MOVEFL_NOGRAVITY))
		{
			V3D_ADDSCALE(cur->vel, world->gravity,
				dt, cur->vel);
		}
		V3D_ADDSCALE(cur->vel, cur->u_acc, dt, cur->vel);
		BSDE_PlaneD_TransformVector(cur->u_racc, cur->xform, dv);
		V3D_ADDSCALE(cur->vel, dv, dt, cur->vel);
	}
}


void BSDE_UpdateVel(BSDE_World *world, BSDE_Solid *cur, double dt)
{
	if(isnan(V3D_DOT(cur->ivel, cur->ivel)))
	{
		V3D_ZERO(cur->ivel);
	}
	if(isnan(V3D_DOT(cur->itvel, cur->itvel)))
	{
		V3D_ZERO(cur->itvel);
	}

#if 0
	if(V3D_LEN(cur->ivel)>10)
	{
		V3D_SCALE(cur->ivel, 10/V3D_LEN(cur->ivel), cur->ivel);
	}
	if((V3D_LEN(cur->itvel)/cur->mass)>10)
	{
		V3D_SCALE(cur->itvel, 10/(V3D_LEN(cur->itvel)/cur->mass),
			cur->itvel);
	}
#endif


	V3D_ADD(cur->vel, cur->ivel, cur->vel);
	V3D_ADD(cur->tvel, cur->itvel, cur->tvel);
//	V3D_ADDSCALE(cur->vel, cur->ivel, dt, cur->vel);
//	V3D_ADDSCALE(cur->e_tvel, cur->itvel, dt, cur->tvel);
	V3D_ZERO(cur->ivel);
	V3D_ZERO(cur->itvel);

#if 0
	if(V3D_LEN(cur->vel)>100)
	{
		V3D_SCALE(cur->vel, 100/V3D_LEN(cur->vel), cur->vel);
	}

	if((V3D_LEN(cur->tvel)/cur->mass)>10)
	{
		V3D_SCALE(cur->tvel, 10/(V3D_LEN(cur->vel)/cur->mass),
			cur->tvel);
	}
#endif

	if((cur->solidtype==BSDE_SOLID_AABB) ||
		(cur->moveflags&BSDE_MOVEFL_NOSPIN))
	{
		BSDE_QuatD_Identity(cur->rot);
		V3D_ZERO(cur->tvel);
	}

	if(cur->moveflags&BSDE_MOVEFL_NOMOVE)
	{
		V3D_ZERO(cur->vel);
	}
}

void BSDE_Update(BSDE_World *world, BSDE_Solid *cur, double dt)
{
	int i;

	V3D_COPY(cur->e_org, cur->org);
	V4D_COPY(cur->e_rot, cur->rot);

#if 0
	//calculate current inertia tensor

	BSDE_WorldInertiaTensor(cur);

	//calculate current transform
	BSDE_QuatD_ToMatrix(cur->rot, cur->xform);
	for(i=0; i<3; i++)cur->xform[3*4+i]+=cur->org[i];
#endif
}

void BSDE_Predict(BSDE_World *world, BSDE_Solid *cur, double dt)
{
	double qa[4], qb[4], n[3], l;
	double ma[9], mb[9], mc[9];
	int i;

	//predict origin and rotation

	V3F_ADDSCALE(cur->org, cur->vel, dt, cur->e_org);


	l=V3D_NORMALIZE(cur->avel, n);

//	printf("spin (%f %f %f) %f\n", n[0], n[1], n[2], l);

	BSDE_QuatD_FromAxis(n, l*dt, qa);
	BSDE_QuatD_Multiply(qa, cur->rot, qb);
//	BSDE_QuatD_Multiply(cur->rot, qa, qb);
	BSDE_QuatD_Normalize(qb, cur->e_rot);

	V3D_COPY(cur->vel, cur->e_vel);
	V3D_COPY(cur->tvel, cur->e_tvel);


	//calculate current inertia tensor

	BSDE_WorldInertiaTensor(cur);

	//calculate current transform
	BSDE_QuatD_ToMatrix(cur->rot, cur->xform);
	for(i=0; i<3; i++)cur->xform[3*4+i]+=cur->org[i];

	//calculate predicted transform
	BSDE_QuatD_ToMatrix(cur->e_rot, cur->e_xform);
	for(i=0; i<3; i++)cur->e_xform[3*4+i]+=cur->e_org[i];


	//calculate move volume
	BSDE_BoxMove(cur, cur->mmins, cur->mmaxs);

	//clear prediction flag (used for expensive internal predictions)
	cur->moveflags&=~BSDE_MOVEFL_PREDICT;
}

void BSDE_PredictPosTime(BSDE_Solid *cur, double dt,
	double *org, double *rot, double *xform)
{
	double qa[4], qb[4], n[3], l;
	int i;

	if(dt>0)
	{
		V3D_ADDSCALE(cur->org, cur->vel, dt, org);

		l=V3D_NORMALIZE(cur->avel, n);
		BSDE_QuatD_FromAxis(n, l*dt, qa);
		BSDE_QuatD_Multiply(qa, cur->rot, qb);
		BSDE_QuatD_Normalize(qb, rot);
	}else
	{
		V3D_COPY(cur->org, org);
		V3D_COPY(cur->rot, rot);
	}

	BSDE_QuatD_ToMatrix(rot, xform);
	for(i=0; i<3; i++)xform[3*4+i]+=org[i];
}

void BSDE_SetWorldGravity(BSDE_World *world, double x, double y, double z)
{
	world->gravity[0]=x;
	world->gravity[1]=y;
	world->gravity[2]=z;
}

void BSDE_GetWorldGravity(BSDE_World *world, double *v)
{
	V3D_COPY(world->gravity, v);
}

double BSDE_GetWorldAttrF(BSDE_World *world, int var)
{
	double f;

	f=0;
	switch(var)
	{
	case BSDE_MINSTEP:
		f=world->minstep;
		break;
	case BSDE_MAXSTEP:
		f=world->maxstep;
		break;
	case BSDE_MAXFRAME:
		f=world->maxframe;
		break;
	case BSDE_REALTIME:
		f=world->realtime;
		break;
	case BSDE_WORLDTIME:
		f=world->worldtime;
		break;
	default:
		break;
	}
	return(f);
}

void BSDE_SetWorldAttrF(BSDE_World *world, int var, double val)
{
	switch(var)
	{
	case BSDE_MINSTEP:
		world->minstep=val;
		break;
	case BSDE_MAXSTEP:
		world->maxstep=val;
		break;
	case BSDE_MAXFRAME:
		world->maxframe=val;
		break;
	default:
		break;
	}
}

int BSDE_GetWorldAttrI(BSDE_World *world, int var)
{
	int i;

	i=0;
	switch(var)
	{
	case BSDE_COLLIDE_STRICT:
		i=(world->flags&BSDE_WORLDFL_STRICT)!=0;
		break;
	case BSDE_FRAME_PASSES:
		i=world->passes;
		break;
	default:
		break;
	}
	return(i);
}

void BSDE_SetWorldAttrI(BSDE_World *world, int var, int val)
{
	switch(var)
	{
	case BSDE_COLLIDE_STRICT:
		if(val)world->flags|=BSDE_WORLDFL_STRICT;
			else world->flags&=~BSDE_WORLDFL_STRICT;
		break;
	case BSDE_FRAME_PASSES:
		world->passes=val;
		break;
	default:
		break;
	}
}

void BSDE_GetWorldAttrFV(BSDE_World *world, int var, float *val, int n)
{
	switch(var)
	{
	case BSDE_GRAVITY:
		val[0]=world->gravity[0];
		val[1]=world->gravity[1];
		val[2]=world->gravity[2];
		break;
	default:
		break;
	}
}

void BSDE_SetWorldAttrFV(BSDE_World *world, int var, float *val, int n)
{
	switch(var)
	{
	case BSDE_GRAVITY:
		world->gravity[0]=val[0];
		world->gravity[1]=val[1];
		world->gravity[2]=val[2];
		break;
	default:
		break;
	}
}

void BSDE_GetWorldAttrDV(BSDE_World *world, int var, double *val, int n)
{
	switch(var)
	{
	case BSDE_GRAVITY:
		val[0]=world->gravity[0];
		val[1]=world->gravity[1];
		val[2]=world->gravity[2];
		break;
	default:
		break;
	}
}

void BSDE_SetWorldAttrDV(BSDE_World *world, int var, double *val, int n)
{
	switch(var)
	{
	case BSDE_GRAVITY:
		world->gravity[0]=val[0];
		world->gravity[1]=val[1];
		world->gravity[2]=val[2];
		break;
	default:
		break;
	}
}

void *BSDE_GetBuffer(BSDE_World *world, int var)
{
}

void BSDE_SetBuffer(BSDE_World *world,
	int var, int type, int num, int stride, void *val)
{
	switch(var)
	{
	case BSDE_CONTACT_SELF:
		world->t_arr_hs=val;
		world->t_str_hs=stride;
		world->t_ty_hs=type;
		world->t_num=num;
		break;
	case BSDE_CONTACT_OTHER:
		world->t_arr_ho=val;
		world->t_str_ho=stride;
		world->t_ty_ho=type;
		world->t_num=num;
		break;
	case BSDE_CONTACT_ORIGIN:
		world->t_arr_o=val;
		world->t_str_o=stride;
		world->t_ty_o=type;
		world->t_num=num;
		break;
	case BSDE_CONTACT_NORMAL:
		world->t_arr_n=val;
		world->t_str_n=stride;
		world->t_ty_n=type;
		world->t_num=num;
		break;
	default:
		break;
	}
}

int BSDE_AddSolidWorld(BSDE_World *world, BSDE_Solid *solid)
{
	BSDE_Solid **lst;
	int i;

	for(i=0; i<world->n_solids; i++)
		if(world->solids[i]==solid)
			return(i);
	for(i=0; i<world->n_solids; i++)
		if(!world->solids[i])
	{
		world->solids[i]=solid;
		solid->handle=i;
		return(i);
	}

	i=world->n_solids+256;
	lst=malloc(i*sizeof(BSDE_Solid *));
	memset(lst, 0, i*sizeof(BSDE_Solid *));

	if(world->n_solids)
	{
		memcpy(lst, world->solids,
			world->n_solids*sizeof(BSDE_Solid *));
		free(world->solids);
	}
	world->solids=lst;
	world->n_solids=i;

	i=BSDE_AddSolidWorld(world, solid);
	return(i);
}

int BSDE_AddJointWorld(BSDE_World *world, BSDE_Joint *joint)
{
	BSDE_Joint **lst;
	int i;

	for(i=0; i<world->n_joints; i++)
		if(world->joints[i]==joint)
			return(i);
	for(i=0; i<world->n_joints; i++)
		if(!world->joints[i])
	{
		world->joints[i]=joint;
		return(i);
	}

	i=world->n_joints+256;
	lst=malloc(i*sizeof(BSDE_Joint *));
	memset(lst, 0, i*sizeof(BSDE_Joint *));

	if(world->n_joints)
	{
		memcpy(lst, world->joints,
			world->n_joints*sizeof(BSDE_Joint *));
		free(world->joints);
	}
	world->joints=lst;
	world->n_joints=i;

	i=BSDE_AddJointWorld(world, joint);
	return(i);
}

BSDE_World *BSDE_NewWorld()
{
	BSDE_World *tmp;

	tmp=malloc(sizeof(BSDE_World));
	memset(tmp, 0, sizeof(BSDE_World));

	tmp->maxstep=0.01;
	tmp->minstep=0.0001;
	tmp->maxframe=0.5;

	tmp->contact=malloc(BSDE_MAX_CONTACTS*sizeof(BSDE_Contact *));

	return(tmp);
}


int *BSDE_GetTriggerArray(BSDE_World *world)
	{return(world->tr_arr);}
int BSDE_GetTriggerArrayPos(BSDE_World *world)
	{return(world->tr_cur);}
void BSDE_ResetTriggerArray(BSDE_World *world)
	{world->tr_cur=0;}

void BSDE_SetTriggerArray(BSDE_World *world, int *arr, int num)
{
	world->tr_arr=arr;
	world->tr_num=num;
	world->tr_cur=0;
}

void BSDE_WorldAddTriggerContact(BSDE_World *world,
	BSDE_Solid *obja, BSDE_Solid *objb)
{
	int ia, ib;
	int i;

	if(!world->tr_arr)return;	//no trigger array

	for(i=0; i<world->n_solids; i++)
		if(world->solids[i]==obja)
			break;
	if(i>=world->n_solids)return;	//obja is not in the world
	ia=i;

	for(i=0; i<world->n_solids; i++)
		if(world->solids[i]==objb)
			break;
	if(i>=world->n_solids)return;	//objb is not in the world
	ib=i;


	for(i=0; i<world->tr_cur; i++)
		if((world->tr_arr[i*2+0]==ia) &&
			(world->tr_arr[i*2+1]==ib))
				break;
	if(i<world->tr_cur)return;	//nothing needs to be done

	world->tr_arr[i*2+0]=ia;
	world->tr_arr[i*2+1]=ib;
	world->tr_cur++;
}
