#include "bsde_internal.h"


//3x3 matrices

void BSDE_Mat3D_MatMult(double *a, double *b, double *c)
{
	c[0]=a[0]*b[0] + a[1]*b[3] + a[2]*b[6];
	c[1]=a[0]*b[1] + a[1]*b[4] + a[2]*b[7];
	c[2]=a[0]*b[2] + a[1]*b[5] + a[2]*b[8];
	c[3]=a[3]*b[0] + a[4]*b[3] + a[5]*b[6];
	c[4]=a[3]*b[1] + a[4]*b[4] + a[5]*b[7];
	c[5]=a[3]*b[2] + a[4]*b[5] + a[5]*b[8];
	c[6]=a[6]*b[0] + a[7]*b[3] + a[8]*b[6];
	c[7]=a[6]*b[1] + a[7]*b[4] + a[8]*b[7];
	c[8]=a[6]*b[2] + a[7]*b[5] + a[8]*b[8];
}

void BSDE_Mat3D_Transpose(double *a, double *b)
{
	b[0]=a[0]; b[1]=a[3]; b[2]=a[6];
	b[3]=a[1]; b[4]=a[4]; b[5]=a[7];
	b[6]=a[2]; b[7]=a[5]; b[8]=a[8];
}

void BSDE_Mat3D_Copy(double *a, double *b)
{
	b[0]=a[0]; b[1]=a[1]; b[2]=a[2];
	b[3]=a[3]; b[4]=a[4]; b[5]=a[5];
	b[6]=a[6]; b[7]=a[7]; b[8]=a[8];
}

int BSDE_Mat3D_Equal(double *a, double *b)
{
	int i;

	i=(b[0]==a[0]) && (b[1]==a[1]) && (b[2]==a[2]) &&
	  (b[3]==a[3]) && (b[4]==a[4]) && (b[5]==a[5]) &&
	  (b[6]==a[6]) && (b[7]==a[7]) && (b[8]==a[8]);
	return(i);
}

void BSDE_Mat3D2F_Copy(double *a, float *b)
{
	b[0]=a[0]; b[1]=a[1]; b[2]=a[2];
	b[3]=a[3]; b[4]=a[4]; b[5]=a[5];
	b[6]=a[6]; b[7]=a[7]; b[8]=a[8];
}

void BSDE_Mat3F2D_Copy(float *a, double *b)
{
	b[0]=a[0]; b[1]=a[1]; b[2]=a[2];
	b[3]=a[3]; b[4]=a[4]; b[5]=a[5];
	b[6]=a[6]; b[7]=a[7]; b[8]=a[8];
}


void BSDE_Mat3D_Scale(double *a, double b, double *c)
{
	c[0]=a[0]*b; c[1]=a[1]*b; c[2]=a[2]*b;
	c[3]=a[3]*b; c[4]=a[4]*b; c[5]=a[5]*b;
	c[6]=a[6]*b; c[7]=a[7]*b; c[8]=a[8]*b;
}


void BSDE_Mat3D_TransformVector(double *dir, double *xform, double *v)
{
	v[0]=(dir[0]*xform[0])+(dir[1]*xform[3])+(dir[2]*xform[6]);
	v[1]=(dir[0]*xform[1])+(dir[1]*xform[4])+(dir[2]*xform[7]);
	v[2]=(dir[0]*xform[2])+(dir[1]*xform[5])+(dir[2]*xform[8]);
}

void BSDE_Mat3D_TransformVectorT(double *dir, double *xform, double *v)
{
	v[0]=(dir[0]*xform[0])+(dir[1]*xform[1])+(dir[2]*xform[2]);
	v[1]=(dir[0]*xform[3])+(dir[1]*xform[4])+(dir[2]*xform[5]);
	v[2]=(dir[0]*xform[6])+(dir[1]*xform[7])+(dir[2]*xform[8]);
}


//4x4 matrices

void BSDE_Mat4D_Copy(double *a, double *b)
{
	b[0]=a[0];   b[1]=a[1];   b[2]=a[2];   b[3]=a[3];
	b[4]=a[4];   b[5]=a[5];   b[6]=a[6];   b[7]=a[7];
	b[8]=a[8];   b[9]=a[9];   b[10]=a[10]; b[11]=a[11];
	b[12]=a[12]; b[13]=a[13]; b[14]=a[14]; b[15]=a[15];
}

int BSDE_Mat4D_Equal(double *a, double *b)
{
	int i;

	i=(b[0] ==a[0]) &&(b[1] ==a[1]) &&(b[2] ==a[2]) &&(b[3] ==a[3]) &&
	  (b[4] ==a[4]) &&(b[5] ==a[5]) &&(b[6] ==a[6]) &&(b[7] ==a[7]) &&
	  (b[8] ==a[8]) &&(b[9] ==a[9]) &&(b[10]==a[10])&&(b[11]==a[11]) &&
	  (b[12]==a[12])&&(b[13]==a[13])&&(b[14]==a[14])&&(b[15]==a[15]);
	return(i);
}

void BSDE_Mat4D_MatMult(double *a, double *b, double *c)
{
	c[0] =a[0]*b[0]  + a[1]*b[4]  + a[2]*b[8]   + a[3]*b[12];
	c[1] =a[0]*b[1]  + a[1]*b[5]  + a[2]*b[9]   + a[3]*b[13];
	c[2] =a[0]*b[2]  + a[1]*b[6]  + a[2]*b[10]  + a[3]*b[14];
	c[3] =a[0]*b[3]  + a[1]*b[7]  + a[2]*b[11]  + a[3]*b[15];
	c[4] =a[4]*b[0]  + a[5]*b[4]  + a[6]*b[8]   + a[7]*b[12];
	c[5] =a[4]*b[1]  + a[5]*b[5]  + a[6]*b[9]   + a[7]*b[13];
	c[6] =a[4]*b[2]  + a[5]*b[6]  + a[6]*b[10]  + a[7]*b[14];
	c[7] =a[4]*b[3]  + a[5]*b[7]  + a[6]*b[11]  + a[7]*b[15];
	c[8] =a[8]*b[0]  + a[9]*b[4]  + a[10]*b[8]  + a[11]*b[12];
	c[9] =a[8]*b[1]  + a[9]*b[5]  + a[10]*b[9]  + a[11]*b[13];
	c[10]=a[8]*b[2]  + a[9]*b[6]  + a[10]*b[10] + a[11]*b[14];
	c[11]=a[8]*b[3]  + a[9]*b[7]  + a[10]*b[11] + a[11]*b[15];
	c[12]=a[12]*b[0] + a[13]*b[4] + a[14]*b[8]  + a[15]*b[12];
	c[13]=a[12]*b[1] + a[13]*b[5] + a[14]*b[9]  + a[15]*b[13];
	c[14]=a[12]*b[2] + a[13]*b[6] + a[14]*b[10] + a[15]*b[14];
	c[15]=a[12]*b[3] + a[13]*b[7] + a[14]*b[11] + a[15]*b[15];
}

void BSDE_Mat4D_Transpose(double *a, double *b)
{
	b[0]=a[0];  b[1]=a[4];  b[2]=a[8];   b[3]=a[12];
	b[4]=a[1];  b[5]=a[5];  b[6]=a[9];   b[7]=a[13];
	b[8]=a[2];  b[9]=a[6];  b[10]=a[10]; b[11]=a[14];
	b[12]=a[3]; b[13]=a[7]; b[14]=a[11]; b[15]=a[15];
}

void BSDE_Mat4D_InverseTransform(double *a, double *b)
{
	b[0]=a[0];	b[1]=a[4];	b[2]=a[8];	b[3]=0;
	b[4]=a[1];	b[5]=a[5];	b[6]=a[9];	b[7]=0;
	b[8]=a[2];	b[9]=a[6];	b[10]=a[10];	b[11]=0;
	b[12]=-a[12];	b[13]=-a[13];	b[14]=-a[14];	b[15]=1;

	b[12]=-V3D_DOT(a+3*4, b+0*4);
	b[13]=-V3D_DOT(a+3*4, b+1*4);
	b[14]=-V3D_DOT(a+3*4, b+2*4);
}


//3x3 matrices inside 3x4 or 4x4 matrices

void BSDE_Mat34D_MatMult(double *a, double *b, double *c)
{
	c[0] =a[0]*b[0] + a[1]*b[4] + a[2]*b[8];
	c[1] =a[0]*b[1] + a[1]*b[5] + a[2]*b[9];
	c[2] =a[0]*b[2] + a[1]*b[6] + a[2]*b[10];
	c[4] =a[4]*b[0] + a[5]*b[4] + a[6]*b[8];
	c[5] =a[4]*b[1] + a[5]*b[5] + a[6]*b[9];
	c[6] =a[4]*b[2] + a[5]*b[6] + a[6]*b[10];
	c[8] =a[8]*b[0] + a[9]*b[4] + a[10]*b[8];
	c[9] =a[8]*b[1] + a[9]*b[5] + a[10]*b[9];
	c[10]=a[8]*b[2] + a[9]*b[6] + a[10]*b[10];
}

void BSDE_Mat34D_Transpose(double *a, double *b)
{
	b[0]=a[0];  b[1]=a[4];  b[2]=a[8];
	b[4]=a[1];  b[5]=a[5];  b[6]=a[9];
	b[8]=a[2];  b[9]=a[6];  b[10]=a[10];
}


// general matrices

void BSDE_MatND_MatMult(double *a, int ah, int aw, double *b, int bh, int bw, double *c)
{
	int i, j, k;

	if(aw!=bh)return;

	for(i=0; i<(ah*bw); i++)c[i]=0;
	for(i=0; i<ah; i++)
		for(j=0; j<bw; j++)
			for(k=0; k<aw; k++)
				c[i*bw+j]+=a[i*aw+k]*b[k*bw+j];
}

void BSDE_MatND_Inverse(double *a, double *b, int n)
{
	static double a2buf[8*8*2], a3buf[16];
	double *a2p[8], *a3p;
	double f;
	int i, j, k;

	for(i=0; i<n*n*2; i++)a2buf[i]=0;
	for(i=0; i<n*n; i++)b[i]=0;

	for(i=0; i<n; i++)a2p[i]=&a2buf[i*n*2];
	for(i=0; i<n; i++)
		for(j=0; j<n; j++)a2p[i][j]=a[i*n+j];
	for(i=0; i<n; i++)a2p[i][i+n]=1.0;

	for(i=0; i<n; i++)
	{
		if(a2p[i][i]==0) /* search for a new pivot in a lower row */
		{
			for(j=n-1; j>i; j--)if(a2p[j][i]!=0)
			{
				a3p=a2p[i];
				a2p[i]=a2p[j];
				a2p[j]=a3p;
			}
			if(a2p[i][i]==0)
			{
				printf("no pivot\n");
				return; /* no pivot */
			}
		}
		f=1.0/a2p[i][i];
		for(j=0; j<2*n; j++)a2p[i][j]*=f;
		for(j=0; j<n; j++)if(i!=j)
		{
			f=-a2p[j][i];
			for(k=0; k<2*n; k++)a3buf[k]=a2p[i][k]*f;
			for(k=0; k<2*n; k++)a2p[j][k]+=a3buf[k];
		}
	}

	for(i=0; i<n; i++)
		for(j=0; j<n; j++)
			b[i*n+j]=a2p[i][n+j];
}
