#include "bsde_internal.h"

void BSDE_PlaneD_LinePlaneIntersect(double *start, double *end,
	double *norm, double *point)
{
	double dir[3], x;

	//calc direction
	V3D_SUB(end, start, dir);

	x=V3D_DOT(dir, norm);
	x=(x<0)?-x:x;
	V3D_SCALE(dir, 1/x, dir);

	//calc intersection
	x=V3D_NDOT(start, norm);
	x=(x<0)?-x:x;
	V3D_ADDSCALE(start, dir, x, point);
}

int BSDE_PlaneD_CheckLinePlane(double *start, double *end, double *norm)
{
	double x;
	int i;

	i=0;
	x=V3D_NDOT(start, norm);
	if(x>=0)i|=1; else i|=2;

	x=V3D_NDOT(end, norm);
	if(x>=0)i|=1; else i|=2;

	return(i);
}

int BSDE_PlaneD_ClipLinePlane(double *start, double *end,
	double *norm, double *end2)
{
	int i;

	i=BSDE_PlaneD_CheckLinePlane(start, end, norm);
	if(i!=3)return(i);

	BSDE_PlaneD_LinePlaneIntersect(start, end, norm, end2);
	return(i);
}

void BSDE_PlaneD_NearestPointLinePlane(double *start, double *end,
	double *norm, double *pt)
{
	double d0, d1;
	int i;

	i=BSDE_PlaneD_CheckLinePlane(start, end, norm);
	if(i==3)
	{
		BSDE_PlaneD_LinePlaneIntersect(start, end, norm, pt);
		return;
	}

	d0=V3D_NDOT(start, norm);
	d1=V3D_NDOT(end, norm);

	if(fabs(d0-d1)<0.001)
	{
		V3D_SCALEADDSCALE(start, 0.5, end, 0.5, pt);
	}else if(fabs(d0)<fabs(d1))
	{
		V3D_COPY(start, pt);
	}else
	{
		V3D_COPY(end, pt);
	}
}

int BSDE_PlaneD_CheckPlaneBox(double *norm, double *mins, double *maxs)
{
	double d0, d1;
	int i;

	d0=0;
	d1=0;
	for(i=0; i<3; i++)
	{
		if(norm[i]>0)
		{
			d0+=maxs[i]*norm[i];
			d1+=mins[i]*norm[i];
		}else
		{
			d0+=mins[i]*norm[i];
			d1+=maxs[i]*norm[i];
		}
	}

	i=0;
	if(d0>=norm[3])i|=1;	//>=dist
	if(d1<norm[3])i|=2;	//<dist

	return(i);
}

void BSDE_PlaneD_TransformVector(double *dir, double *xform, double *v)
{
	v[0]=(dir[0]*xform[0])+(dir[1]*xform[4])+(dir[2]*xform[8]);
	v[1]=(dir[0]*xform[1])+(dir[1]*xform[5])+(dir[2]*xform[9]);
	v[2]=(dir[0]*xform[2])+(dir[1]*xform[6])+(dir[2]*xform[10]);
}

void BSDE_PlaneD_TransformVectorT(double *dir, double *xform, double *v)
{
	v[0]=(dir[0]*xform[0])+(dir[1]*xform[1])+(dir[2]*xform[2]);
	v[1]=(dir[0]*xform[4])+(dir[1]*xform[5])+(dir[2]*xform[6]);
	v[2]=(dir[0]*xform[8])+(dir[1]*xform[9])+(dir[2]*xform[10]);
}

void BSDE_PlaneD_TransformNormal(double *dir, double *xform, double *norm)
{
	double f;

	BSDE_PlaneD_TransformVector(dir, xform, norm);
	f=(norm[0]*xform[12])+(norm[1]*xform[13])+(norm[2]*xform[14]);
	norm[3]=dir[3]+f;
}

void BSDE_PlaneD_TransformNormalT(double *dir, double *xform, double *norm)
{
	double f;

	BSDE_PlaneD_TransformVectorT(dir, xform, norm);
	f=(norm[0]*xform[12])+(norm[1]*xform[13])+(norm[2]*xform[14]);
	norm[3]=dir[3]+f;
}

void BSDE_PlaneD_TransformNormalN(double *dir, double *xform, double *norm)
{
	double f;

	BSDE_PlaneD_TransformVector(dir, xform, norm);
	f=(norm[0]*xform[12])+(norm[1]*xform[13])+(norm[2]*xform[14]);
	norm[3]=dir[3]-f;
}

void BSDE_PlaneD_TransformNormalNT(double *dir, double *xform, double *norm)
{
	double f;

	BSDE_PlaneD_TransformVectorT(dir, xform, norm);
	f=(norm[0]*xform[12])+(norm[1]*xform[13])+(norm[2]*xform[14]);
	norm[3]=dir[3]-f;
}

void BSDE_PlaneD_TransformPoint(double *point, double *xform, double *org)
{
	double f;

	org[0]=(point[0]*xform[0])+(point[1]*xform[4])+
		(point[2]*xform[8])+xform[12];
	org[1]=(point[0]*xform[1])+(point[1]*xform[5])+
		(point[2]*xform[9])+xform[13];
	org[2]=(point[0]*xform[2])+(point[1]*xform[6])+
		(point[2]*xform[10])+xform[14];
}

void BSDE_PlaneD_TransformPointT(double *point, double *xform, double *org)
{
	double f;

	org[0]=(point[0]*xform[0])+(point[1]*xform[1])+
		(point[2]*xform[2])+xform[12];
	org[1]=(point[0]*xform[4])+(point[1]*xform[5])+
		(point[2]*xform[6])+xform[13];
	org[2]=(point[0]*xform[8])+(point[1]*xform[9])+
		(point[2]*xform[10])+xform[14];
}

void BSDE_PlaneD_TransformPointN(double *point, double *xform, double *org)
{
	double f;

	org[0]=(point[0]*xform[0])+(point[1]*xform[4])+
		(point[2]*xform[8])-xform[12];
	org[1]=(point[0]*xform[1])+(point[1]*xform[5])+
		(point[2]*xform[9])-xform[13];
	org[2]=(point[0]*xform[2])+(point[1]*xform[6])+
		(point[2]*xform[10])-xform[14];
}

void BSDE_PlaneD_TransformPointNT(double *point, double *xform, double *org)
{
	double f;

	org[0]=(point[0]*xform[0])+(point[1]*xform[1])+
		(point[2]*xform[2])-xform[12];
	org[1]=(point[0]*xform[4])+(point[1]*xform[5])+
		(point[2]*xform[6])-xform[13];
	org[2]=(point[0]*xform[8])+(point[1]*xform[9])+
		(point[2]*xform[10])-xform[14];
}

void BSDE_PlaneD_TransformPointLocal(double *point, double *xform, double *org)
{
	double pt[3], f;

	V3D_SUB(pt, xform+3*4, pt);
	org[0]=(pt[0]*xform[0])+(pt[1]*xform[1])+(pt[2]*xform[2]);
	org[1]=(pt[0]*xform[4])+(pt[1]*xform[5])+(pt[2]*xform[6]);
	org[2]=(pt[0]*xform[8])+(pt[1]*xform[9])+(pt[2]*xform[10]);

#if 0
	BSDE_PlaneD_TransformVectorT(xform+3*4, xform, dv);
	org[0]= (point[0]*xform[0]) +(point[1]*xform[1])+
		(point[2]*xform[2]) -dv[0];
	org[1]= (point[0]*xform[4]) +(point[1]*xform[5])+
		(point[2]*xform[6]) -dv[1];
	org[2]= (point[0]*xform[8]) +(point[1]*xform[9])+
		(point[2]*xform[10])-dv[2];
#endif
}

void BSDE_PlaneD_AdjacentNormals(double *n, double *x, double *y)
{
	double f, g;

	V3D_ZERO(x);
	V3D_ZERO(y);

	if(fabs(n[0])>fabs(n[1]))
	{
		if(fabs(n[0])>fabs(n[2]))
		{
			x[1]=1;
			y[2]=1;
		}else
		{
			x[0]=1;
			y[1]=1;
		}
	}else
	{
		if(fabs(n[1])>fabs(n[2]))
		{
			x[0]=1;
			y[2]=1;
		}else
		{
			x[0]=1;
			y[1]=1;
		}
	}

	f=V3D_DOT(x, n);
	g=V3D_DOT(y, n);
	V3D_ADDSCALE(x, n, -f, x);
	V3D_ADDSCALE(y, n, -g, y);

	V3D_NORMALIZE(x, x);
	V3D_NORMALIZE(y, y);
}
