#include "bsde_internal.h"

//Generic AABB stuff

void BSDE_SolidAABB_BoxMove(double *bmins, double *bmaxs,
	double *start, double *end, double *mins, double *maxs)
{
	double mins2[3], maxs2[3];
	int i;

	V3D_ADD(bmins, start, mins);
	V3D_ADD(bmaxs, start, maxs);
	V3D_ADD(bmins, end, mins2);
	V3D_ADD(bmaxs, end, maxs2);

	for(i=0; i<3; i++)
	{
		if(mins2[i]<mins[i])mins[i]=mins2[i];
		if(maxs2[i]>maxs[i])maxs[i]=maxs2[i];
	}
}

int BSDE_SolidAABB_BoxCollideP(double *ma, double *na, double *mb, double *nb)
{
	if(na[0]<=mb[0])return(0);
	if(na[1]<=mb[1])return(0);
	if(na[2]<=mb[2])return(0);
	if(nb[0]<=ma[0])return(0);
	if(nb[1]<=ma[1])return(0);
	if(nb[2]<=ma[2])return(0);

	return(1);
}

int BSDE_SolidAABB_BoxCollideAxis(double *ma, double *na, double *mb, double *nb)
{
	int i;

	i=0;
	if((na[0]>mb[0]) && !(ma[0]>=nb[0]))i|=1;
	if((na[1]>mb[1]) && !(ma[1]>=nb[1]))i|=2;
	if((na[2]>mb[2]) && !(ma[2]>=nb[2]))i|=4;

	if((nb[0]>ma[0]) && !(mb[0]>=na[0]))i|=8;
	if((nb[1]>ma[1]) && !(mb[1]>=na[1]))i|=16;
	if((nb[2]>ma[2]) && !(mb[2]>=na[2]))i|=32;

	return(i);
}

int BSDE_SolidAABB_PointInBoxP(double *p, double *min, double *max)
{
	if(p[0]<=min[0])return(0);
	if(p[0]>=max[0])return(0);
	if(p[1]<=min[1])return(0);
	if(p[1]>=max[1])return(0);
	if(p[2]<=min[2])return(0);
	if(p[2]>=max[2])return(0);

	return(1);
}

void BSDE_SolidAABB_BoxIntersection(double *ma, double *na,
	double *mb, double *nb, double *mc, double *nc)
{
	mc[0]=(ma[0]>mb[0])?ma[0]:mb[0];
	nc[0]=(na[0]<nb[0])?na[0]:nb[0];
	mc[1]=(ma[1]>mb[1])?ma[1]:mb[1];
	nc[1]=(na[1]<nb[1])?na[1]:nb[1];
	mc[2]=(ma[2]>mb[2])?ma[2]:mb[2];
	nc[2]=(na[2]<nb[2])?na[2]:nb[2];
}

int BSDE_SolidAABB_BoxCollideContact(double *ma, double *na,
	double *mb, double *nb, double *org, double *norm, double *dist)
{
	double mc[3], nc[3];
	double bd, d;
	int i, bi;

	BSDE_SolidAABB_BoxIntersection(ma, na, mb, nb, mc, nc);
	bd=99999999.0;
	bi=-1;

	for(i=0; i<3; i++)
	{
		d=nc[i]-mc[i];
		if(d<0)return(0);	//does not collide
//		if(d<bd)
		if((0<d) && (d<bd))
		{
			bi=i;
			bd=d;
		}
	}
	if(bi<0)
	{
		bi=2;
		bd=0;
	}

	org[0]=(mc[0]+nc[0])*0.5;
	org[1]=(mc[1]+nc[1])*0.5;
	org[2]=(mc[2]+nc[2])*0.5;

	V3D_ZERO(norm);
	i=(ma[bi]<mb[bi])?-1:1;
	norm[bi]=i;
	*dist=bd;

	return(1);
}

void BSDE_SolidAABB_BoxPlaneExtents(double *mins, double *maxs,
	double *norm, double *min, double *max)
{
	double d0, d1;
	int i;

	d0=0;
	d1=0;
	for(i=0; i<3; i++)
	{
		if(norm[i]>0)
		{
			d0+=maxs[i]*norm[i];
			d1+=mins[i]*norm[i];
		}else
		{
			d0+=mins[i]*norm[i];
			d1+=maxs[i]*norm[i];
		}
	}

	*min=d1;
	*max=d0;
}
