#include "bsde_internal.h"


//BREP

void BSDE_SolidBRep_ProjectTriangleLine(
	double *u0, double *u1, double *u2,
	double *iv, double *rm, double *rn)
{
	double d0, d1, d2;
	double m0, n0;

	d0=V3D_DOT(iv, u0);
	d1=V3D_DOT(iv, u1);
	d2=V3D_DOT(iv, u2);
	m0=(d0<d1)?((d0<d2)?d0:d2):((d1<d2)?d1:d2);
	n0=(d0>d1)?((d0>d2)?d0:d2):((d1>d2)?d1:d2);

	*rm=m0;
	*rn=n0;
}

//check if 2 triangles overlap on a particular line
int BSDE_SolidBRep_CheckLineTriangleTriangle(
	double *u0, double *u1, double *u2,
	double *v0, double *v1, double *v2,
	double *iv)
{
	double d0, d1, d2;
	double m0, n0, m1, n1;

	d0=V3D_DOT(iv, u0);
	d1=V3D_DOT(iv, u1);
	d2=V3D_DOT(iv, u2);
	m0=(d0<d1)?((d0<d2)?d0:d2):((d1<d2)?d1:d2);
	n0=(d0>d1)?((d0>d2)?d0:d2):((d1>d2)?d1:d2);

	d0=V3D_DOT(iv, v0);
	d1=V3D_DOT(iv, v1);
	d2=V3D_DOT(iv, v2);
	m1=(d0<d1)?((d0<d2)?d0:d2):((d1<d2)?d1:d2);
	n1=(d0>d1)?((d0>d2)?d0:d2):((d1>d2)?d1:d2);

	if((m0>n1) || (m1>n0))
		return(0);
	return(1);
}

//check if 2 triangles overlap on a particular line
//adjusts pt to fall within the space
int BSDE_SolidBRep_CheckLineTriangleTriangle2(
	double *u0, double *u1, double *u2,
	double *v0, double *v1, double *v2,
	double *iv, double *pt)
{
	double d0, d1, d2;
	double m0, n0, m1, n1, m2, n2;

	d0=V3D_DOT(iv, u0);
	d1=V3D_DOT(iv, u1);
	d2=V3D_DOT(iv, u2);
	m0=(d0<d1)?((d0<d2)?d0:d2):((d1<d2)?d1:d2);
	n0=(d0>d1)?((d0>d2)?d0:d2):((d1>d2)?d1:d2);

	d0=V3D_DOT(iv, v0);
	d1=V3D_DOT(iv, v1);
	d2=V3D_DOT(iv, v2);
	m1=(d0<d1)?((d0<d2)?d0:d2):((d1<d2)?d1:d2);
	n1=(d0>d1)?((d0>d2)?d0:d2):((d1>d2)?d1:d2);

	if((m0>n1) || (m1>n0))
		return(0);

	m2=(m0>m1)?m0:m1;
	n2=(n0<n1)?n0:n1;

	d0=V3D_DOT(iv, pt);
	if(d0<m2)V3D_ADDSCALE(pt, iv, m2-d0, pt);
	if(d0>n2)V3D_ADDSCALE(pt, iv, n2-d0, pt);

	return(1);
}


//check if 2 triangles overlap
int BSDE_SolidBRep_CheckTriangleTriangle(
	double *u0, double *u1, double *u2,
	double *v0, double *v1, double *v2)
{
	double du0[3], du1[3], du2[4], du3[3];
	double dv0[3], dv1[3], dv2[4], dv3[3];
	double iv[3];
	double d0, d1, d2, f;
	int i;


#if 1
	//check: v does not cross u

	V3D_SUB(u1, u0, du0);
	V3D_SUB(u2, u0, du1);
	V3D_CROSS(du0, du1, du2);
	du2[3]=V3D_DOT(du2, u0);

	d0=V3D_DOT(du2, v0)-du2[3];
	d1=V3D_DOT(du2, v1)-du2[3];
	d2=V3D_DOT(du2, v2)-du2[3];

	if(((d0*d1)>0) && ((d0*d2)>0))
		return(0);
#endif

#if 1
	//check: u does not cross v

	V3D_SUB(v1, v0, dv0);
	V3D_SUB(v2, v0, dv1);
	V3D_CROSS(dv0, dv1, dv2);
	dv2[3]=V3D_DOT(dv2, v0);

	d0=V3D_DOT(dv2, u0)-dv2[3];
	d1=V3D_DOT(dv2, u1)-dv2[3];
	d2=V3D_DOT(dv2, u2)-dv2[3];

	if(((d0*d1)>0) && ((d0*d2)>0))
		return(0);
#endif

#if 1
	//joint overlap

	V3D_CROSS(du2, dv2, iv);
	f=V3D_NORMALIZE(iv, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);
#endif

#if 1
	//check sides
	V3D_CROSS(du2, du0, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);

	V3D_CROSS(du2, du1, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);

#if 1
	V3D_SUB(u2, u1, du3);
	V3D_CROSS(du2, du3, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);
#endif
#endif

#if 1
	V3D_CROSS(dv2, dv0, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);

	V3D_CROSS(dv2, dv1, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);


#if 1
	//finally, check the C side
	V3D_SUB(v2, v1, dv3);
	V3D_CROSS(dv2, dv3, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);
#endif
#endif

	return(1);
}

//check if 2 triangles overlap also while computing intersections
int BSDE_SolidBRep_CheckTriangleTriangle2(
	double *u0, double *u1, double *u2,
	double *v0, double *v1, double *v2,
	double *iorg, double *idir)
{
	double du0[3], du1[3], du2[4], du3[3];
	double dv0[3], dv1[3], dv2[4], dv3[3];
	double iv[3], pt[3];
	double d0, d1, d2;
	double f;
	int i;

	V3D_ZERO(pt);

#if 1
	//check: v does not cross u

	V3D_SUB(u1, u0, du0);
	V3D_SUB(u2, u0, du1);
	V3D_CROSS(du0, du1, du2);
	du2[3]=V3D_DOT(du2, u0);

	d0=V3D_DOT(du2, v0)-du2[3];
	d1=V3D_DOT(du2, v1)-du2[3];
	d2=V3D_DOT(du2, v2)-du2[3];

	if(((d0*d1)>0) && ((d0*d2)>0))
		return(0);

	f=V3D_DOT(pt, du2)-du2[3];
	V3F_ADDSCALE(pt, du2, -f, pt);
#endif

#if 1
	//check: u does not cross v

	V3D_SUB(v1, v0, dv0);
	V3D_SUB(v2, v0, dv1);
	V3D_CROSS(dv0, dv1, dv2);
	dv2[3]=V3D_DOT(dv2, v0);

	d0=V3D_DOT(dv2, u0)-dv2[3];
	d1=V3D_DOT(dv2, u1)-dv2[3];
	d2=V3D_DOT(dv2, u2)-dv2[3];

	if(((d0*d1)>0) && ((d0*d2)>0))
		return(0);

	f=V3D_DOT(pt, dv2)-dv2[3];
	V3D_ADDSCALE(pt, dv2, -f, pt);
#endif

#if 1
	//joint overlap

	V3D_CROSS(du2, dv2, iv);
	f=V3D_NORMALIZE(iv, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle(
		u0, u1, u2, v0, v1, v2, iv);
	if(!i)return(0);
#endif

#if 1
	//check sides
	V3D_CROSS(du2, du0, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);

	V3D_CROSS(du2, du1, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);

#if 1
	V3D_SUB(u2, u1, du3);
	V3D_CROSS(du2, du3, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);
#endif
#endif

#if 1
	V3D_CROSS(dv2, dv0, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);

	V3D_CROSS(dv2, dv1, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);


#if 1
	//finally, check the C side
	V3D_SUB(v2, v1, dv3);
	V3D_CROSS(dv2, dv3, iv);
	i=BSDE_SolidBRep_CheckLineTriangleTriangle2(
		u0, u1, u2, v0, v1, v2, iv, pt);
	if(!i)return(0);
#endif
#endif

	V3D_COPY(pt, iorg);

	return(1);
}

int BSDE_SolidBRep_GetTriangleNormal(
	double *u0, double *u1, double *u2,
	double *io)
{
	double du0[3], du1[3];

	V3D_SUB(u1, u0, du0);
	V3D_SUB(u2, u0, du1);
	V3D_CROSS(du0, du1, io);

	return(0);
}

#if 0
int BSDE_SolidBRep_GetTriangleIntersection(
	double *u0, double *u1, double *u2,
	double *v0, double *v1, double *v2,
	double *io)
{
	double du0[3], du1[3], du2[4];
	double dv0[3], dv1[3], dv2[4];
	double pt[3];
	double f;
	int i;

	V3D_ZERO(pt);

	V3D_SUB(u1, u0, du0);
	V3D_SUB(u2, u0, du1);
	V3D_CROSS(du0, du1, du2);
	V3D_Normalize(du2, du2);
	du2[3]=V3D_DOT(du2, u0);

	f=V3D_DOT(pt, du2)-du2[3];
	VecNF_ADDSCALE(pt, du2, -f, pt, 3);

	V3D_SUB(v1, v0, dv0);
	V3D_SUB(v2, v0, dv1);
	V3D_CROSS(dv0, dv1, dv2);
	V3D_Normalize(dv2, dv2);
	dv2[3]=V3D_DOT(dv2, v0);

	f=V3D_DOT(pt, dv2)-dv2[3];
	VecNF_ADDSCALE(pt, dv2, -f, pt, 3);

	return(0);
}

#endif

//check 2 transformed brep solids
//av: vertices, at: triangles, na: count
int BSDE_SolidBRep_CheckBrep(
	double *av, int *at, int na,
	double *bv, int *bt, int nb,
	double *rio, double *riv)
{
	double io[3], iv[3];
	double *u0, *u1, *u2;
	double *v0, *v1, *v2;
	double w0[3], w1[3], w2[3];
	double m0, n0, apd, f;
	int i, j, k, l, li;

	V3D_ZERO(io);
	V3D_ZERO(iv);
	apd=0;

	l=0;
	for(i=0; i<(na-1); i++)
		for(j=0; j<(nb-1); j++)
	{
		u0=av+(at[i*3+0]*3);
		u1=av+(at[i*3+1]*3);
		u2=av+(at[i*3+2]*3);
		v0=bv+(bt[j*3+0]*3);
		v1=bv+(bt[j*3+1]*3);
		v2=bv+(bt[j*3+2]*3);

		k=BSDE_SolidBRep_CheckTriangleTriangle(
			u0, u1, u2, v0, v1, v2);

		if(k && rio)
		{
			w0[0]=u0[0]+u1[0]+u2[0];
			w0[1]=u0[1]+u1[1]+u2[1];
			w0[2]=u0[2]+u1[2]+u2[2];
			w1[0]=v0[0]+v1[0]+v2[0];
			w1[1]=v0[1]+v1[1]+v2[1];
			w1[2]=v0[2]+v1[2]+v2[2];

			io[0]+=w0[0]+w1[0];
			io[1]+=w0[1]+w1[1];
			io[2]+=w0[2]+w1[2];
		}

		if(k && (rio && riv))
		{
			BSDE_SolidBRep_GetTriangleNormal(u0, u1, u2, w2);
			iv[0]+=w2[0];
			iv[1]+=w2[1];
			iv[2]+=w2[2];

			BSDE_SolidBRep_ProjectTriangleLine(
				v0, v1, v2, w2, &m0, &n0);
			f=-(m0-V3D_DOT(u0, w2));
			apd+=f;
		}

		if(k)l++;
	}

	if(rio)
	{
		li=l?l:1;

		rio[0]=io[0]/(6*li);
		rio[1]=io[1]/(6*li);
		rio[2]=io[2]/(6*li);
	}

	if(riv)
	{
		V3D_NORMALIZE(iv, riv);
		riv[3]=apd;
	}

	return(l);
}

//check 2 transformed brep solids
//av: vertices, at: triangles, na: count
//rio is the returned collision points, padded to 4 floats
//riv is the returned collision normals and penetration distances
//rn is the limit for the number of returned points
//returns the number of collision points
int BSDE_SolidBRep_CheckBrep2(
	double *av, int *at, int na,
	double *bv, int *bt, int nb,
	double *rio, double *riv, int rn)
{
	double io[3], iv[3];
	double *u0, *u1, *u2;
	double *v0, *v1, *v2;
	double w0[3], w1[3], w2[3];
	double m0, n0, apd, f;
	int i, j, k, l, li;

	V3D_ZERO(io);
	V3D_ZERO(iv);
	apd=0;

	l=0;
	for(i=0; i<(na-1); i++)
		for(j=0; j<(nb-1); j++)
	{
		u0=av+(at[i*3+0]*3);
		u1=av+(at[i*3+1]*3);
		u2=av+(at[i*3+2]*3);
		v0=bv+(bt[j*3+0]*3);
		v1=bv+(bt[j*3+1]*3);
		v2=bv+(bt[j*3+2]*3);

		k=BSDE_SolidBRep_CheckTriangleTriangle(
			u0, u1, u2, v0, v1, v2);

		if(k && rio)
		{
			w0[0]=u0[0]+u1[0]+u2[0];
			w0[1]=u0[1]+u1[1]+u2[1];
			w0[2]=u0[2]+u1[2]+u2[2];
			w1[0]=v0[0]+v1[0]+v2[0];
			w1[1]=v0[1]+v1[1]+v2[1];
			w1[2]=v0[2]+v1[2]+v2[2];

			rio[0]=(w0[0]+w1[0])/6;
			rio[1]=(w0[1]+w1[1])/6;
			rio[2]=(w0[2]+w1[2])/6;
			rio[3]=0;
			rio+=4;

			if(riv)
			{
				BSDE_SolidBRep_GetTriangleNormal(u0, u1, u2, w2);
				riv[0]=w2[0];
				riv[1]=w2[1];
				riv[2]=w2[2];

				BSDE_SolidBRep_ProjectTriangleLine(
					v0, v1, v2, w2, &m0, &n0);
				f=-(m0-V3D_DOT(u0, w2));
				riv[3]=f;

				riv+=4;
			}
		}

		if(k)l++;
	}

	if(rio)
	{
		li=l?l:1;

		rio[0]=io[0]/(6*li);
		rio[1]=io[1]/(6*li);
		rio[2]=io[2]/(6*li);
	}

	if(riv)
	{
		V3D_NORMALIZE(iv, riv);
		riv[3]=apd;
	}

	return(l);
}

int BSDE_SolidBRep_TrimBrep(double *av, int *at, int na,
	int *at2, double *bmins, double *bmaxs)
{
	double m[3], n[3];
	double *v0, *v1, *v2;
	int i, j, k;

	j=0;
	for(i=0; i<na; i++)
	{
		v0=av+(at[i*3+0]*3);
		v1=av+(at[i*3+1]*3);
		v2=av+(at[i*3+2]*3);

		for(k=0; k<3; k++)
		{
			m[k]=(v0[k]<v1[k])?
				((v0[k]<v2[k])?v0[k]:v2[k]):
				((v1[k]<v2[k])?v1[k]:v2[k]);
			n[k]=(v0[k]>v1[k])?
				((v0[k]>v2[k])?v0[k]:v2[k]):
				((v1[k]>v2[k])?v1[k]:v2[k]);
		}

		if(BSDE_SolidAABB_BoxCollideP(m, n, bmins, bmaxs))
		{
			at2[j*3+0]=at[i*3+0];
			at2[j*3+1]=at[i*3+1];
			at2[j*3+2]=at[i*3+2];
			j++;
		}
	}
	return(j);
}


void BSDE_SolidBRep_ProjectLine(
	double *vecs, int *tris, int nv, int nt,
	double *pos, double *norm, double *rmin, double *rmax)
{
	double tv[3], m, n, f;
	int i;


	m=BSDE_M_BIGVAL;
	n=-BSDE_M_BIGVAL;
	for(i=0; i<nv; i++)
	{
		BSDE_PlaneD_TransformPoint(vecs+i*3, pos, tv);
		f=V3D_DOT(tv, norm);
		if(f<m)m=f;
		if(f>n)n=f;
	}

	*rmin=m;
	*rmax=n;
}

void BSDE_SolidBRep_ProjectVecsLine(
	double *vecs, int nv,
	double *norm, double *rmin, double *rmax)
{
	double tv[3], m, n, f;
	int i;

	m=BSDE_M_BIGVAL; n=-BSDE_M_BIGVAL;
	for(i=0; i<nv; i++)
	{
		f=V3D_DOT(vecs+i*3, norm);
		if(f<m)m=f;
		if(f>n)n=f;
	}

	*rmin=m; *rmax=n;
}

void BSDE_SolidBRep_ProjectLine2(BSDE_Solid *obj, double *pos,
	double *norm, double *rmin, double *rmax)
{
	//use predicted info if possible
	if((obj->moveflags&BSDE_MOVEFL_PREDICT) &&
		BSDE_Mat4D_Equal(pos, obj->e_xform))
	{
		BSDE_SolidBRep_ProjectVecsLine(
			obj->e_vecs, obj->n_vecs, norm, rmin, rmax);
		return;
	}

	BSDE_SolidBRep_ProjectLine(obj->vecs, obj->faces,
		obj->n_vecs, obj->n_faces, pos, norm, rmin, rmax);
}

//predict line in ending pos
//use cached values if available
void BSDE_SolidBRep_ProjectLine3(BSDE_Solid *obj,
	double *norm, double *rmin, double *rmax)
{
	int i;

	if(obj->moveflags&BSDE_MOVEFL_PREDICT)
	{
		BSDE_SolidBRep_ProjectVecsLine(
			obj->e_vecs, obj->n_vecs, norm, rmin, rmax);
		return;
	}

	BSDE_SolidBRep_ProjectLine(obj->vecs, obj->faces,
		obj->n_vecs, obj->n_faces, obj->e_xform, norm, rmin, rmax);
}

//use a few teqniques to attempt to find an approximate for the volume
//this function will assume a "solid" mesh
double BSDE_SolidBRep_ApproxBrepVolume(double *vecs, int *tris, int nv, int nt)
{
	double mins[3], maxs[3], org[3], org1[3];
	double mins2[3], maxs2[3];
	double tv[3], iv0[3], iv1[3], iv2[3];
	double sr0, sr1, f, vol;
	int i;

	//AABB
	V3D_ZERO(org);
	V3D_INITBOX(mins, maxs);
	for(i=0; i<nv; i++)
	{
		V3D_COPY(vecs+i*3, tv);
		V3D_ADD(org, tv, org);
		V3D_MIN(mins, tv, mins);
		V3D_MAX(maxs, tv, maxs);
	}
	V3D_SCALE(org, 1.0/nv, org);
	vol=(maxs[0]-mins[0])*(maxs[1]-mins[1])*(maxs[2]-mins[2]);

	org1[0]=(mins[0]+maxs[0])*0.5;
	org1[1]=(mins[1]+maxs[1])*0.5;
	org1[2]=(mins[2]+maxs[2])*0.5;

	//Spheres
	sr0=0; sr1=0;
	for(i=0; i<nv; i++)
	{
		//from est COM
		f=V3D_DIST(vecs+i*3, org);
		if(f>sr0)sr0=f;

		//from est center
		f=V3D_DIST(vecs+i*3, org1);
		if(f>sr1)sr1=f;
	}

	f=4.0/3.0*BSDE_M_PI*sr0*sr0;
	if(f<vol)vol=f;
	f=4.0/3.0*BSDE_M_PI*sr1*sr1;
	if(f<vol)vol=f;

	//OBB
	V3D_ZERO(iv0);
	V3D_ZERO(iv1);

	for(i=0; i<nv; i++)
	{
		iv0[0]+=fabs(vecs[i*3+0]-org[0]);
		iv0[1]+=fabs(vecs[i*3+0]-org[1]);
		iv0[2]+=fabs(vecs[i*3+0]-org[2]);
	}
	V3D_NORMALIZE(iv0, iv0);

	for(i=0; i<nv; i++)
	{
		V3D_SUB(vecs+i*3, org, tv);
		f=V3D_DOT(tv, iv0);
		V3D_ADDSCALE(tv, iv0, -f, tv);

		iv1[0]+=fabs(tv[0]);
		iv1[1]+=fabs(tv[1]);
		iv1[2]+=fabs(tv[2]);
	}
	V3D_NORMALIZE(iv1, iv1);

	V3D_CROSS(iv0, iv1, iv2);

	V3D_INITBOX(mins2, maxs2);
	for(i=0; i<nv; i++)
	{
		tv[0]=V3D_DOT(vecs+i*3, iv0);
		tv[1]=V3D_DOT(vecs+i*3, iv1);
		tv[2]=V3D_DOT(vecs+i*3, iv2);
		V3D_MIN(mins2, tv, mins2);
		V3D_MAX(maxs2, tv, maxs2);
	}

	f=(maxs2[0]-mins2[0])*(maxs2[1]-mins2[1])*(maxs2[2]-mins2[2]);
	if(f<vol)vol=f;

	return(vol);
}

void BSDE_SolidBRep_PlaneIntersect(BSDE_Solid *obj, double *pos,
	double *norm, double *io)
{
	double bo[3], pt[3];
	double d;
	int i, j;

	V3D_ZERO(bo);
	j=0;

	for(i=0; i<obj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(obj->vecs+i*3, pos, pt);
		d=V3D_NDOT(pt, norm);
		if(d<0)
		{
			V3D_ADD(bo, pt, bo);
			j++;
		}
	}

	if(!j)j++;
	V3D_SCALE(bo, 1.0/j, io);
}

void BSDE_SolidBRep_Predict(BSDE_Solid *obj)
{
	int i;

	if(obj->moveflags&BSDE_MOVEFL_PREDICT)
		return;

	if(!obj->e_vecs)obj->e_vecs=
		malloc(obj->n_vecs*3*sizeof(double));
	if(!obj->e_evecs)obj->e_evecs=
		malloc(obj->n_faces*12*sizeof(double));

	for(i=0; i<obj->n_vecs; i++)
		BSDE_PlaneD_TransformPoint(
			obj->vecs+i*3, obj->e_xform, obj->e_vecs+i*3);

#if 0
	for(i=0; i<(obj->n_faces*3); i++)
		BSDE_PlaneD_TransformNormal(
			obj->evecs+i*4, obj->e_xform, obj->e_evecs+i*4);
#endif

	obj->moveflags|=BSDE_MOVEFL_PREDICT;
}

int BSDE_SolidBRep_CheckCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *apos, double *bpos)
{
	int i, j;

	i=0;
	switch(bobj->solidtype)
	{
	case BSDE_SOLID_AABB:
	case BSDE_SOLID_OBB:
		break;

	case BSDE_SOLID_SPHERE:
		break;

	case BSDE_SOLID_CYLINDER:
	case BSDE_SOLID_CCYLINDER:
		break;

	case BSDE_SOLID_PLANE:
		break;

	case BSDE_SOLID_HULL:
		i=BSDE_SolidHull_CheckCollide(bobj, aobj, bpos, apos);
		break;
	case BSDE_SOLID_TRIMESH:
//		i=BSDE_SolidHull_CheckTriMeshCollide(
//			aobj, bobj, axform, bxform);
		break;

	default:
		break;
	}

	return(i);
}

int BSDE_SolidBRep_MultiContact(BSDE_World *world,
	BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *apos, double *bpos, BSDE_Contact **lst, int num)
{
	double io[3], iv[3], id;
	int i, j;

//	printf("brep\n");

	i=BSDE_SolidAABB_BoxCollideP(aobj->mmins, aobj->mmaxs,
		bobj->mmins, bobj->mmaxs);
	if(!i)return(num);

	switch(bobj->solidtype)
	{
	case BSDE_SOLID_AABB:
	case BSDE_SOLID_OBB:
		break;

	case BSDE_SOLID_SPHERE:
		break;

	case BSDE_SOLID_CYLINDER:
	case BSDE_SOLID_CCYLINDER:
		break;

	case BSDE_SOLID_PLANE:
		break;

	case BSDE_SOLID_HULL:
		i=BSDE_SolidHull_CheckTriMeshCollide(bobj, aobj, bpos, apos);
		if(i)
		{
			BSDE_SolidHull_ContactGeneric(bobj, aobj,
				bpos, apos, bpos+3*4, apos+3*4, io, iv, &id);
			lst[num++]=BSDE_MakeContact(world, bobj, aobj, io, iv, id);
		}
		break;
	case BSDE_SOLID_TRIMESH:
//		i=BSDE_SolidHull_CheckTriMeshCollide(
//			aobj, bobj, axform, bxform);
		break;

	default:
		break;
	}

	return(num);
}
