#include "bsde_internal.h"

//Cylinders
//Issue: may have issues in some cases, eg, long/narrow cylinders, or
//parallel+offset cylinders

void BSDE_SolidCylinder_MakeBox(double height, double radius, int t,
	double *pos, double *mins, double *maxs)
{
	double f;
	int i;

	for(i=0; i<3; i++)
	{
		if(t)
		{
			f=fabs(pos[2*4+i]*height*0.5)+radius;
			mins[i]=pos[3*4+i]-f;
			maxs[i]=pos[3*4+i]+f;
		}else
		{
			f=	fabs(pos[0*4+i]*radius)+
				fabs(pos[1*4+i]*radius)+
				fabs(pos[2*4+i]*height*0.5);
			mins[i]=pos[3*4+i]-f;
			maxs[i]=pos[3*4+i]+f;
		}
	}
}

void BSDE_SolidCylinder_BoxMove(
	double height, double radius, int t,
	double *spos, double *epos,
	double *mins, double *maxs)
{
	double amins[3], amaxs[3];
	double bmins[3], bmaxs[3];
	int i;

	BSDE_SolidCylinder_MakeBox(height, radius, t, spos, amins, amaxs);
	BSDE_SolidCylinder_MakeBox(height, radius, t, epos, bmins, bmaxs);
	for(i=0; i<3; i++)mins[i]=(amins[i]<bmins[i])?amins[i]:bmins[i];
	for(i=0; i<3; i++)maxs[i]=(amaxs[i]>bmaxs[i])?amaxs[i]:bmaxs[i];
}

#if 0
int BSDE_SolidCylinder_CollideObbContact(
	double alen, double arad, int t, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double pt[3], dv[3], aorg[3];
	double d, r;

	aorg[0]=apos[3];
	aorg[1]=apos[7];
	aorg[2]=apos[11];

	BSDE_SolidOBB_ObbNearestPoint(bmins, bmaxs, bpos, aorg, pt);
	V3D_SUB(aorg, pt, dv);
	d=V3D_NORMALIZE(dv, dv);

	if(t)
	{
		r=arad +fabs(V3D_DOT(apos+(2*4), dv)*alen*0.5);
	}else
	{
		r=	fabs(V3D_DOT(apos+(0*4), dv)*arad)+
			fabs(V3D_DOT(apos+(1*4), dv)*arad)+
			fabs(V3D_DOT(apos+(2*4), dv)*alen*0.5);
	}

	if(d>r)return(0);

	if(org)
	{
		V3D_COPY(pt, org);
		V3D_COPY(dv, norm);
		*dist=r-d;
	}

	return(1);
}
#endif

double BSDE_SolidCylinder_AxisRadius(
	double len, double rad, int t, double *pos, double *norm)
{
	double r;

	if(t)
	{
		r=rad +	fabs(V3D_DOT(pos+(2*4), norm)*len*0.5);
	}else
	{
		r=	fabs(V3D_DOT(pos+(0*4), norm)*rad) +
			fabs(V3D_DOT(pos+(1*4), norm)*rad) +
			fabs(V3D_DOT(pos+(2*4), norm)*len*0.5);
	}
	return(r);
}

void BSDE_SolidCylinder_PointLocal(double *pos, double *org, double *pt)
{
	double v[3];

	V3D_SUB(org, pos+3*4, v);
	pt[0]=V3D_DOT(pos+0*4, v);
	pt[1]=V3D_DOT(pos+1*4, v);
	pt[2]=V3D_DOT(pos+2*4, v);
}

void BSDE_SolidCylinder_NearestPoint(
	double len, double rad, int t, double *pos,
	double *point, double *pt)
{
	double v[3], v1[3], f, g;

	V3D_SUB(point, pos+3*4, v1);
	v[0]=V3D_DOT(pos+0*4, v1);
	v[1]=V3D_DOT(pos+1*4, v1);
	v[2]=V3D_DOT(pos+2*4, v1);

	f=V2D_LEN(v);
	if(f>rad)
	{
		V2D_SCALE(v, rad/f, v);
	}

	if(t)
	{
		f=fabs(v[2])-(len*0.5);
		if(f>0)
		{
			g=sqrt(v[0]*v[0]+v[1]*v[1]+f*f);
			if(g>rad)
			{
				V2D_COPY(v, v1);
				v1[2]=(v[2]>0)?f:-f;

				V2D_SCALE(v1, rad/g, v);
			}
		}
	}else
	{
		if(v[2]>(len*0.5))v[2]=len*0.5;
		if(v[2]<(-(len*0.5)))v[2]=-len*0.5;
	}

	pt[0]=v[0]*pos[0]+v[1]*pos[4]+v[2]*pos[8]+pos[12];
	pt[1]=v[0]*pos[1]+v[1]*pos[5]+v[2]*pos[9]+pos[13];
	pt[2]=v[0]*pos[2]+v[1]*pos[6]+v[2]*pos[10]+pos[14];
}

int BSDE_SolidCylinder_PointCollide(
	double len, double rad, int t, double *pos,
	double *point)
{
	double v[3], pt[3], f, g;

	V3D_SUB(point, pos+3*4, pt);
	v[0]=V3D_DOT(pos+0*4, pt);
	v[1]=V3D_DOT(pos+1*4, pt);
	v[2]=V3D_DOT(pos+2*4, pt);

	if(V2F_LEN(v)>=rad)return(0);

	if(t)
	{
		if(fabs(v[2])>=(len*0.5+rad))return(0);

		f=fabs(v[2])-(len*0.5);
		if(f<=0)return(1);
		g=sqrt(v[0]*v[0]+v[1]*v[1]+f*f);
		if(g>=rad)return(0);
	}else
	{
		if(fabs(v[2])>=(len*0.5))return(0);
	}
	return(1);
}

void BSDE_SolidCylinder_OffsetCPA(
	double *ua, double *va, double *ub, double *vb,
	double *da, double *db)
{
	double dp[3];
	double vx, q0, q1, d;

	V3D_SUB(ub, ua, dp);
	q0= V3D_DOT(dp, va);
	q1=-V3D_DOT(dp, vb);

	vx=V3D_DOT(va, vb);
	d=1-vx*vx;

	if(d<=0.0001)
	{
//		*da=0;
//		*db=0;

		*da=V3D_DOT(ub, va)-V3D_DOT(ua, va);
		*db=V3D_DOT(ua, vb)-V3D_DOT(ub, vb);
		return;
	}

	d=1.0/d;
	*da=(q0+vx*q1)*d;
	*db=(vx*q0+q1)*d;
}

int BSDE_SolidCylinder_CollideContact(
	double alen, double arad, int at, double *apos,
	double blen, double brad, int bt, double *bpos,
	double *org, double *norm, double *dist)
{
	double dv[3], aorg[3], borg[3];
	double d, f, g, ra, rb;

	V3D_COPY(apos+3*4, aorg);
	V3D_COPY(bpos+3*4, borg);

	f=V3D_DOT(apos+2*4, bpos+2*4);
	if(fabs(f)>0.9)
	{
		//ok, cylinders are near parallel
		V3D_COPY(apos+2*4, dv);
		ra=V3D_DOT(aorg, dv);
		rb=V3D_DOT(borg, dv);
		if(rb>ra)
		{
			V3D_SCALE(dv, -1, dv);
			ra=V3D_DOT(aorg, dv);
			rb=V3D_DOT(borg, dv);
		}

		f=-alen*0.5;
		g=blen*0.5;

		dv[0]=(dv[0]*f)-(dv[0]*g)+(aorg[0]-borg[0]);
		dv[1]=(dv[1]*f)-(dv[1]*g)+(aorg[1]-borg[1]);
		dv[2]=(dv[2]*f)-(dv[2]*g)+(aorg[2]-borg[2]);
		d=V3D_NORMALIZE(dv, dv);
	}else
	{
		BSDE_SolidCylinder_OffsetCPA(aorg, apos+2*4, borg, bpos+2*4,
			&ra, &rb);
		if(ra>(alen*0.5))ra=alen*0.5;
		if(ra<(-alen*0.5))ra=-alen*0.5;
		if(rb>(blen*0.5))rb=blen*0.5;
		if(rb<(-blen*0.5))rb=-blen*0.5;

		dv[0]=(apos[2*4+0]*ra)-(bpos[2*4+0]*rb)+(aorg[0]-borg[0]);
		dv[1]=(apos[2*4+1]*ra)-(bpos[2*4+1]*rb)+(aorg[1]-borg[1]);
		dv[2]=(apos[2*4+2]*ra)-(bpos[2*4+2]*rb)+(aorg[2]-borg[2]);
		d=V3D_NORMALIZE(dv, dv);
	}

	if(at)ra=arad;
		else ra=fabs(V3D_DOT(apos+0*4, dv)*arad)+
			fabs(V3D_DOT(apos+1*4, dv)*arad);
	if(bt)rb=brad;
		else rb=fabs(V3D_DOT(bpos+0*4, dv)*brad)+
			fabs(V3D_DOT(bpos+1*4, dv)*brad);
	if(d>=(ra+rb))return(0);

	if(org)
	{
		f=rb+(d-ra);
		V3D_ADDSCALE(borg, dv, f, org);

		V3D_COPY(dv, norm);
		*dist=(ra+rb)-d;
	}

	return(1);
}

int BSDE_SolidCylinder_CollideObbContact(
	double alen, double arad, int ty, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double aorg[3];
	double pt[3], dv[3];
	double ext[4];
	double bn[3];
	double d0, d1, d, r, bd;
	int i, j;

	aorg[0]=apos[12]; aorg[1]=apos[13]; aorg[2]=apos[14];

	BSDE_SolidOBB_ObbNearestPoint(bmins, bmaxs, bpos, aorg, pt);
	V3D_SUB(aorg, pt, dv);
	d=V3D_NORMALIZE(dv, dv);
	if(d<0.0001)
	{
		dv[2]=1;
		pt[2]-=arad;
	}

	r=BSDE_SolidCylinder_AxisRadius(alen, arad, ty, apos, dv);
	if(d>r)return(0);

	V3D_COPY(dv, bn);
	bd=r-d;

	for(i=0; i<3; i++)
	{
		r=BSDE_SolidCylinder_AxisRadius(alen, arad, ty, apos,
			bpos+i*4);
		ext[0]=V3D_DOT(aorg, bpos+i*4)-r;
		ext[1]=V3D_DOT(aorg, bpos+i*4)+r;

		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<=0)return(0);

		if(d<bd)
		{
			V3D_COPY(bpos+i*4, bn);
			bd=d;
		}
	}

	if(org)
	{
		V3D_COPY(pt, org);
		V3D_NORMALIZE(bn, bn);
		V3D_COPY(bn, norm);
		*dist=bd;
	}
	return(1);
}
