#include "bsde_internal.h"

//OBB

int BSDE_SolidOBB_CheckObbPlane(double *mins, double *maxs,
	double *xform, double *norm, double *io)
{
	double w[3];

	double xwa, ywa, zwa;
	double m, n;
	int i;

	w[0]=V3D_DOT(xform+(0*4), norm);
	w[1]=V3D_DOT(xform+(1*4), norm);
	w[2]=V3D_DOT(xform+(2*4), norm);

	xwa=xform[3*4+0]*norm[0];
	ywa=xform[3*4+1]*norm[1];
	zwa=xform[3*4+2]*norm[2];

	m=0;
	n=0;
	for(i=0; i<3; i++)
	{
		if(w[i]>0)
		{
			m+=mins[i]*w[i];
			n+=maxs[i]*w[i];
		}else
		{
			m+=maxs[i]*w[i];
			n+=mins[i]*w[i];
		}
	}
	m+=xwa+ywa+zwa-norm[3];
	n+=xwa+ywa+zwa-norm[3];

	i=0;
	if(m<0)i|=2;
	if(n>=0)i|=1;

	if(!i)
	{
		printf("BSDE_SolidOBB_CheckObbPlane: fail\n");

		printf("[(%f %f %f) (%f %f %f)]\n",
			mins[0], mins[1], mins[2],
			maxs[0], maxs[1], maxs[2]);

		printf("(%f %f %f %f)\n",
			norm[0], norm[1], norm[2], norm[3]);

		printf("%f %f\n", m, n);
	}

	if(io && (i==3))
		BSDE_SolidOBB_ObbPlaneIntersect(
			mins, maxs, xform, norm, io);

	return(i);
}

#if 0
int BSDE_SolidOBB_ObbPlaneIntersect(double *mins, double *maxs,
	double *xform, double *norm, double *io)
{
	double dx[3], dy[3], dz[3], dorg[3], dc[3];

	dc[0]=(mins[0]+maxs[0])*0.5;
	dc[1]=(mins[1]+maxs[1])*0.5;
	dc[2]=(mins[2]+maxs[2])*0.5;

	dorg[0]=xform[3*4+0];
	dorg[1]=xform[3*4+1];
	dorg[2]=xform[3*4+2];
	BSDE_Vec3D_ProjectPointOnPlane2(dorg, dorg, norm, norm[3]);

	BSDE_Vec3D_ProjectPointOnPlane(dx, xform+(0*4), norm);
	BSDE_Vec3D_ProjectPointOnPlane(dy, xform+(1*4), norm);
	BSDE_Vec3D_ProjectPointOnPlane(dz, xform+(2*4), norm);


	io[0]=dorg[0]+(dc[0]*dx[0])+(dc[1]*dy[0])+(dc[2]*dz[0]);
	io[1]=dorg[1]+(dc[0]*dx[1])+(dc[1]*dy[1])+(dc[2]*dz[1]);
	io[2]=dorg[2]+(dc[0]*dx[2])+(dc[1]*dy[2])+(dc[2]*dz[2]);

	return(0);
}
#endif

int BSDE_SolidOBB_ObbPlaneIntersect(double *mins, double *maxs,
	double *xform, double *norm, double *io)
{
	double pts[8*3];
	double bo[3];
	double d;
	int i, j;

	BSDE_SolidOBB_GenPoints(mins, maxs, xform, pts);

	V3D_ZERO(bo);
	j=0;

	for(i=0; i<8; i++)
	{
		d=V3D_NDOT(pts+i*3, norm);
		if(d<0)
		{
			V3D_ADD(bo, pts+i*3, bo);
			j++;
		}
	}

	if(!j)j++;
	V3D_SCALE(bo, 1.0/j, io);

	return(0);
}


int BSDE_SolidOBB_ObbPlaneExtents(double *mins, double *maxs,
	double *xform, double *norm, double *ext)
{
	double w[3];

	double xwa, ywa, zwa;
	double m, n;
	int i;

	w[0]=V3D_DOT(xform+(0*4), norm);
	w[1]=V3D_DOT(xform+(1*4), norm);
	w[2]=V3D_DOT(xform+(2*4), norm);

	xwa=xform[3*4+0]*norm[0];
	ywa=xform[3*4+1]*norm[1];
	zwa=xform[3*4+2]*norm[2];

//	m=999999999;
//	n=-999999999;
	m=0;
	n=0;

	for(i=0; i<3; i++)
	{
		if(w[i]>0)
		{
			m+=mins[i]*w[i];
			n+=maxs[i]*w[i];
		}else
		{
			m+=maxs[i]*w[i];
			n+=mins[i]*w[i];
		}
	}
	m+=xwa+ywa+zwa;
	n+=xwa+ywa+zwa;

	ext[0]=m;
	ext[1]=n;

	return(0);
}

void BSDE_SolidOBB_PlaneExtents(double *mins, double *maxs,
	double *xform, double *norm, double *min, double *max)
{
	double w[3];
	double m, n, d;
	int i;

	w[0]=V3D_DOT(xform+(0*4), norm);
	w[1]=V3D_DOT(xform+(1*4), norm);
	w[2]=V3D_DOT(xform+(2*4), norm);
	d=V3D_DOT(xform+3*4, norm);

	m=0;
	n=0;
	for(i=0; i<3; i++)
	{
		if(w[i]>0)
		{
			m+=mins[i]*w[i];
			n+=maxs[i]*w[i];
		}else
		{
			m+=maxs[i]*w[i];
			n+=mins[i]*w[i];
		}
	}
	*min=m+d;
	*max=n+d;
}

int BSDE_SolidOBB_ObbPlaneMoveExtents(double *mins, double *maxs,
	double *sxform, double *exform, double *norm, double *ext)
{
	double ext2[4];

	BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, sxform, norm, ext2+0);
	BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, exform, norm, ext2+2);

	ext[0]=(ext2[0]<ext2[2])?ext2[0]:ext2[2];
	ext[1]=(ext2[1]>ext2[3])?ext2[1]:ext2[3];

	return(0);
}

int BSDE_SolidOBB_CheckObbMovePlane(double *mins, double *maxs,
	double *sxform, double *exform, double *norm, double *io)
{
	double ext[2];
	int i;

	BSDE_SolidOBB_ObbPlaneMoveExtents(mins, maxs, sxform, exform,
		norm, ext);

	i=0;
	if((ext[0]-norm[3])<0)i|=2;
	if((ext[1]-norm[3])>=0)i|=1;

	if(io && (i==3))
		BSDE_SolidOBB_ObbPlaneIntersect(
			mins, maxs, exform, norm, io);

	return(i);
}

int BSDE_SolidOBB_CheckCollide(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos)
{
	int i, j;
	double ext[8], dir[3];
	double omins[3], omaxs[3];

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			apos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			apos+i*4, ext+2);
		if((ext[0]>=ext[3]) || (ext[2]>=ext[1]))
			return(0);
	}

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bpos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		if((ext[0]>=ext[3]) || (ext[2]>=ext[1]))
			return(0);

		omins[i]=(ext[0]>ext[2])?ext[0]:ext[2];	//greater min
		omaxs[i]=(ext[1]<ext[3])?ext[1]:ext[3];	//lesser max
	}

	for(i=0; i<3; i++)
		for(j=0; j<3; j++)
	{
		V3D_CROSS(apos+i*4, bpos+j*4, dir);

		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bpos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		if((ext[0]>=ext[3]) || (ext[2]>=ext[1]))
			return(0);
	}

	return(1);
}

int BSDE_SolidOBB_CheckObbMove(
	double *amins, double *amaxs, double *aspos, double *aepos,
	double *bmins, double *bmaxs, double *bspos, double *bepos,
	double *org)
{
	int i;
	double ext[8];
	double omins[3], omaxs[3], iorg[3];

#if 0
	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, aspos,
			aspos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bspos,
			aspos+i*4, ext+4);
	}
#endif

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, aepos,
			aepos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bepos,
			aepos+i*4, ext+2);
		if((ext[0]>ext[3]) || (ext[2]>ext[1]))
			return(0);
	}

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, aepos,
			bepos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bepos,
			bepos+i*4, ext+2);
		if((ext[0]>ext[3]) || (ext[2]>ext[1]))
			return(0);

		omins[i]=(ext[0]>ext[2])?ext[0]:ext[2];	//greater min
		omaxs[i]=(ext[1]<ext[3])?ext[1]:ext[3];	//lesser max
	}


	for(i=0; i<3; i++)
		iorg[i]=(omins[i]+omaxs[i])*0.5;
	for(i=0; i<3; i++)
	{
		org[i]=iorg[0]*bepos[0*4+i]+
			iorg[1]*bepos[1*4+i]+
			iorg[2]*bepos[2*4+i]+
			bepos[3*4+i];
	}

	return(1);
}

int BSDE_SolidOBB_ObbImpactNormal(
	double *amins, double *amaxs, double *aspos, double *aepos,
	double *bmins, double *bmaxs, double *bspos, double *bepos,
	double *vel, double *norm)
{
	double rv[3], rn[3];
	double ext[2];
	int i;

	rv[0]=V3D_DOT(vel, bepos+0*4);
	rv[1]=V3D_DOT(vel, bepos+1*4);
	rv[2]=V3D_DOT(vel, bepos+2*4);

	if(fabs(rv[0])>fabs(rv[1]))
	{
		if(fabs(rv[0])>fabs(rv[2]))
		{
			rn[0]=(rv[0]>0)?1:-1;
			rn[1]=0;
			rn[2]=0;
		}else
		{
			rn[0]=0;
			rn[1]=0;
			rn[2]=(rv[2]>0)?1:-1;
		}
	}else
	{
		if(fabs(rv[1])>fabs(rv[2]))
		{
			rn[0]=0;
			rn[1]=(rv[1]>0)?1:-1;
			rn[2]=0;
		}else
		{
			rn[0]=0;
			rn[1]=0;
			rn[2]=(rv[2]>0)?1:-1;
		}
	}

	for(i=0; i<3; i++)
	{
		norm[i]=rn[0]*bepos[0*4+i]+
			rn[1]*bepos[1*4+i]+
			rn[2]*bepos[2*4+i];
	}

	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bepos,
		norm, ext);
	norm[3]=ext[0];	//near extent

	return(0);
}

void BSDE_SolidOBB_Box(double *mins, double *maxs, double *pos,
	double *bmins, double *bmaxs)
{
	double norm[3];
	double ext[2];
	int i;

	for(i=0; i<3; i++)
	{
		V3D_ZERO(norm);
		norm[i]=1;

		BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, pos,
			norm, ext);
		bmins[i]=ext[0];
		bmaxs[i]=ext[1];
	}
}

void BSDE_SolidOBB_BoxMove(
	double *mins, double *maxs, double *spos, double *epos,
	double *bmins, double *bmaxs)
{
	double norm[3];
	double ext[2];
	int i;

	for(i=0; i<3; i++)
	{
		V3D_ZERO(norm);
		norm[i]=1;

		BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, spos,
			norm, ext);
		bmins[i]=ext[0];
		bmaxs[i]=ext[1];

		BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, epos,
			norm, ext);
		if(ext[0]<bmins[i])bmins[i]=ext[0];
		if(ext[1]>bmaxs[i])bmaxs[i]=ext[1];
	}
}

#if 0
int BSDE_SolidOBB_CollideContact(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double ext[4];
	double dv[3], aorg[3], borg[3];
	double bn[3], v0[3], v1[3];
	double d0, d1, d, bd, f;
	int i, j;

	V3D_ZERO(bn);
	V3D_ZERO(org);
	bd=99999999.0;
//	bd=-99999999.0;

//	aorg[0]=apos[3]; aorg[1]=apos[7]; aorg[2]=apos[11];
//	borg[0]=bpos[3]; borg[1]=bpos[7]; borg[2]=bpos[11];

	V3D_COPY(apos+12, aorg);
	V3D_COPY(bpos+12, borg);

	BSDE_SolidOBB_ObbNearestPoint(amins, amaxs, apos, borg, v0);
	BSDE_SolidOBB_ObbNearestPoint(bmins, bmaxs, bpos, aorg, v1);

	V3D_SUB(v1, v0, dv);

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			apos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			apos+i*4, ext+2);
//		if((ext[0]>ext[3]) || (ext[2]>ext[1]))
//			return(0);

		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);

//		f=(V3D_DOT(apos+i*4, dv)>0)?1:-1;
		f=(V3D_DOT(aorg, apos+i*4)>V3D_DOT(borg, apos+i*4))?1:-1;
		if(d<bd)
		{
			V3D_SCALE(apos+i*4, f, bn);
			bd=d;
		}

		d=V3D_DOT(org, apos+i*4);
		V3D_ADDSCALE(org, apos+i*4, ((d0+d1)*0.5)-d, org);
	}

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bpos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);

//		f=(V3D_DOT(bpos+i*4, dv)>0)?1:-1;
		f=(V3D_DOT(aorg, bpos+i*4)>V3D_DOT(borg, bpos+i*4))?1:-1;
		if(d<bd)
		{
			V3D_SCALE(apos+i*4, f, bn);
			bd=d;
		}

		d=V3D_DOT(org, bpos+i*4);
		V3D_ADDSCALE(org, bpos+i*4, ((d0+d1)*0.5)-d, org);
	}

//	if(j)bd=0;
//	V3D_SCALE(bn, bo, org);

//	V3D_SCALE(org, 1.0/3.0, org);

//	BSDE_SolidOBB_ObbBestNormalPoint(bmins, bmaxs, bpos,
//		org, bn, &bd);

#if 0
	BSDE_SolidOBB_ObbBestNormalPointDir(amins, amaxs, apos,
		org, dv, v0, &d0);
	BSDE_SolidOBB_ObbBestNormalPointDir(bmins, bmaxs, bpos,
		org, dv, v1, &d1);
	if(d0<d1)
	{
//		V3D_SCALE(v0, -1, bn);
		V3D_COPY(v0, bn);
		bd=d0;
	}else
	{
		V3D_COPY(v1, bn);
		bd=d1;
	}
#endif

//	bn[0]=org[0]-bpos[4];
//	bn[1]=org[1]-bpos[8];
//	bn[2]=org[2]-bpos[12];

	V3D_NORMALIZE(bn, bn);
	V3D_COPY(bn, norm);
	*dist=bd;
	return(1);
}
#endif

void BSDE_SolidOBB_ContactPoint(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org)
{
	double ext[4];
	double dv[3], aorg[3], borg[3];
	double bn[3], v0[3], v1[3];
	double d0, d1, d, bd, f;
	int i, j;

	V3D_ZERO(org);

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			apos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			apos+i*4, ext+2);

		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=V3D_DOT(org, apos+i*4);
		V3D_ADDSCALE(org, apos+i*4, ((d0+d1)*0.5)-d, org);
	}

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bpos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=V3D_DOT(org, bpos+i*4);
		V3D_ADDSCALE(org, bpos+i*4, ((d0+d1)*0.5)-d, org);
	}
}

void BSDE_SolidOBB_ObbNearestPoint(
	double *mins, double *maxs, double *pos,
	double *org, double *pt)
{
	double n[3], o[3];
	double d;
	int i;

	V3D_COPY(org, pt);
//	o[0]=pos[3]; o[1]=pos[7]; o[2]=pos[11];
	V3D_COPY(pos+12, o);

	for(i=0; i<3; i++)
	{
		V3D_COPY(pos+(i*4), n);
		d=V3D_DOT(pt, n)-V3D_DOT(o, n);
		if(d<mins[i])
		{
			V3D_ADDSCALE(pt, n, mins[i]-d, pt);
		}
		if(d>maxs[i])
		{
			V3D_ADDSCALE(pt, n, maxs[i]-d, pt);
		}
	}
}

#if 1
void BSDE_SolidOBB_ObbBestNormalPoint(
	double *mins, double *maxs, double *pos,
	double *org, double *norm, double *dist)
{
	double bn[3];
	double n[3], o[3];
	double d, f, bd;
	int i;

//	o[0]=pos[3]; o[1]=pos[7]; o[2]=pos[11];
	V3D_COPY(pos+12, o);

	V3D_ZERO(bn);
	bd=99999999.0;

	for(i=0; i<3; i++)
	{
		V3D_COPY(pos+(i*4), n);
		d=V3D_DOT(org, n)-V3D_DOT(o, n);

		f=d-mins[i];
		if((f<bd) && (f>0))
//		if(f<bd)
		{
			V3D_SCALE(n, -1, bn);
			bd=f;
		}

		f=maxs[i]-d;
		if((f<bd) && (f>0))
//		if(f<bd)
		{
			V3D_COPY(n, bn);
			bd=f;
		}
	}

	V3D_COPY(bn, norm);
	*dist=bd;
}
#endif

double BSDE_SolidOBB_ObbOverlapDist(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *norm)
{
	double ext[4];
	double d0, d1;

	BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, norm, ext+0);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, norm, ext+2);
	d0=(ext[0]>ext[2])?ext[0]:ext[2];
	d1=(ext[1]<ext[3])?ext[1]:ext[3];

	return(d1-d0);
}

double BSDE_SolidOBB_PointOverlapDist(
	double *mins, double *maxs, double *pos,
	double *org, double *norm)
{
	double ext[2];
	double d0, d1;

	BSDE_SolidOBB_ObbPlaneExtents(mins, maxs, pos, norm, ext);
	d1=ext[1];
	d0=V3D_DOT(org, norm);

	return(d1-d0);
}

#if 0
void BSDE_SolidOBB_ObbBestNormalPoint(
	double *mins, double *maxs, double *pos,
	double *org, double *norm, double *dist)
{
	double bn[3];
	double n[3], o[3];
	double d, f, bd;
	int i;

//	o[0]=pos[3]; o[1]=pos[7]; o[2]=pos[11];
	V3D_COPY(pos+12, o);

	V3D_ZERO(bn);
	bd=99999999.0;

	for(i=0; i<3; i++)
	{
		V3D_COPY(pos+(i*4), n);
		d=V3D_DOT(org, n)-V3D_DOT(o, n);

		f=d-mins[i];
		V3D_SCALE(n, -1/f, bn);

		f=maxs[i]-d;
		V3D_SCALE(n, 1/f, bn);
	}

	V3D_NORMALIZE(bn, norm);
	bd=BSDE_SolidOBB_PointOverlapDist(mins, maxs, pos, org, norm);

	*dist=bd;
}
#endif

void BSDE_SolidOBB_ObbBestNormalPointDir(
	double *mins, double *maxs, double *pos,
	double *org, double *dir,
	double *norm, double *dist)
{
	double bn[3];
	double n[3], o[3];
	double d, f, g, bd;
	int i;

//	o[0]=pos[3]; o[1]=pos[7]; o[2]=pos[11];
	V3D_COPY(pos+12, o);

	V3D_ZERO(bn);
	bd=99999999.0;

	for(i=0; i<3; i++)
	{
		V3D_COPY(pos+(i*4), n);
		d=V3D_DOT(org, n)-V3D_DOT(o, n);
		g=(V3D_DOT(n, dir)>0)?1:-1;

		f=d-mins[i];
//		if(f<bd)
		if((f<bd) && (f>=0))
		{
			V3D_SCALE(n, g, bn);
			bd=f;
		}

		f=maxs[i]-d;
//		if(f<bd)
		if((f<bd) && (f>=0))
		{
			V3D_SCALE(n, g, bn);
			bd=f;
		}
	}

	V3D_COPY(bn, norm);
	*dist=bd;
}

#if 1
//attempt to find best normal based on faces
int BSDE_SolidOBB_ObbBestNormalObb(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *norm, double *dist)
{
	double ext[4];
	double dv[3], aorg[3], borg[3];
	double bn[3], n[3], v0[3], v1[3];
	double d0, d1, d, bd, f;
	int i, j;

	V3D_ZERO(bn);
	bd=99999999.0;

//	aorg[0]=apos[3]; aorg[1]=apos[7]; aorg[2]=apos[11];
//	borg[0]=bpos[3]; borg[1]=bpos[7]; borg[2]=bpos[11];

	V3D_COPY(apos+12, aorg);
	V3D_COPY(bpos+12, borg);

	j=0;
	for(i=0; i<3; i++)
	{
		V3D_COPY(apos+(i*4), n);
		f=V3D_DOT(aorg, n);

		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, n, ext);
		d0=ext[0]-f;
		d1=ext[1]-f;

		if((d0<amaxs[i]) && (d1>amaxs[i]) && (d0>amins[i]))
//		if((d0<amaxs[i]) && (d1>amaxs[i]))
		{
			d=amaxs[i]-d0;

			V3D_ADD(bn, n, bn);

#if 0
			if((d<bd) || !j)
			{
				V3D_COPY(n, bn);
				bd=d;
				j=1;
			}
#endif
		}

		if((d0<amins[i]) && (d1>amins[i]) && (d1<amaxs[i]))
//		if((d0<amins[i]) && (d1>amins[i]))
		{
			d=d1-amins[i];

			V3D_ADDSCALE(bn, n, -1, bn);

#if 0
			if((d<bd) || !j)
			{
//				V3D_COPY(n, bn);
				V3D_SCALE(n, -1, bn);
				bd=d;
				j=1;
			}
#endif
		}
	}
//	if(!j)return(0);	//no good normals

	V3D_NORMALIZE(bn, bn);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, bn, ext);
	bd=ext[1]-ext[0];

	V3D_COPY(bn, norm);
	*dist=bd;
	return(1);
}
#endif

#if 0
int BSDE_SolidOBB_CollideContact(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double ext[4];
	double bn[3], v0[3], v1[3];
	double d0, d1;
	int i, j;

	i=BSDE_SolidOBB_CheckCollide(
		amins, amaxs, apos, bmins, bmaxs, bpos);
	if(!i)return(0);

	i=BSDE_SolidOBB_ObbBestNormalObb(
		amins, amaxs, apos, bmins, bmaxs, bpos, v0, &d0);

	j=BSDE_SolidOBB_ObbBestNormalObb(
		bmins, bmaxs, bpos, amins, amaxs, apos, v1, &d1);

	BSDE_SolidOBB_ContactPoint(
		amins, amaxs, apos, bmins, bmaxs, bpos, org);


#if 1
	V3D_ADDSCALE(v1, v0, -1, bn);
//	V3D_ADDSCALE(v1, v0, 1, bn);
	V3D_NORMALIZE(bn, bn);

//	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, bn, ext);
//	d0=ext[1]-ext[0];

	BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, bn, ext+0);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, bn, ext+2);
	d0=(ext[0]>ext[2])?ext[0]:ext[2];
	d1=(ext[1]<ext[3])?ext[1]:ext[3];

	V3D_SCALE(bn, 1, norm);

	*dist=d1-d0;
//	*dist=0;
	return(1);
#endif

	if(i && j)
	{
		if(d0<d1)
		{
			V3D_SCALE(v0, -1, norm);
			*dist=d0;
		}else
		{
			V3D_SCALE(v1, 1, norm);
			*dist=d1;
		}
		return(1);
	}

	if(!i && !j)
	{
		printf("normal fail\n");
		return(0);
	}

	if(i)
	{
		V3D_SCALE(v0, -1, norm);
		*dist=d0;
	}else
	{
//		V3D_COPY(p1, org);
		V3D_SCALE(v1, 1, norm);
		*dist=d1;
	}
	return(1);
}
#endif


int BSDE_SolidOBB_ObbPointCollide(
	double *mins, double *maxs, double *pos, double *org)
{
	double d;
	int i;

	for(i=0; i<3; i++)
	{
		d=V3D_DOT(org, pos+(i*4))-pos[3*4+i];
		if(d<mins[i])return(0);
		if(d>maxs[i])return(0);
	}
	return(1);
}

void BSDE_SolidOBB_GenPoints(
	double *mins, double *maxs, double *pos, double *pts)
{
	double v[3];
	int i, n;

#if 0
	for(i=0; i<8; i++)
	{
		V3D_SCALE(pos+0*4, (i&1)?1:-1, pts+i*3);
		V3D_ADDSCALE(pts+i*3, pos+1*4, (i&2)?1:-1, pts+i*3);
		V3D_ADDSCALE(pts+i*3, pos+2*4, (i&4)?1:-1, pts+i*3);
		V3D_ADD(pts+i*3, pos+3*4, pts+i*3);
	}
#endif

	for(i=0; i<8; i++)
	{
		v[0]=(i&1)?maxs[0]:mins[0];
		v[1]=(i&2)?maxs[1]:mins[1];
		v[2]=(i&4)?maxs[2]:mins[2];
		BSDE_PlaneD_TransformPoint(v, pos, pts+i*3);
	}
}

void BSDE_SolidOBB_GenPoints4x(
	double *mins, double *maxs, double *pos, double *pts)
{
	double v[3], d[3];
	int i, j, k, n;

	d[0]=(maxs[0]-mins[0])/3.0;
	d[1]=(maxs[1]-mins[1])/3.0;
	d[2]=(maxs[2]-mins[2])/3.0;

	for(i=0; i<4; i++)
		for(j=0; j<4; j++)
			for(k=0; k<4; k++)
	{
		v[0]=mins[0]+d[0]*i;
		v[1]=mins[1]+d[1]*j;
		v[2]=mins[2]+d[2]*k;
		BSDE_PlaneD_TransformPoint(v, pos, pts+(i*16+j*4+k)*3);
	}
}

void BSDE_SolidOBB_GenPointsNx(
	double *mins, double *maxs, double *pos, double *pts, int n)
{
	double v[3], d[3];
	int i, j, k, n2;

	d[0]=(maxs[0]-mins[0])/(n-1);
	d[1]=(maxs[1]-mins[1])/(n-1);
	d[2]=(maxs[2]-mins[2])/(n-1);

	n2=n*n;
	for(i=0; i<n; i++)
		for(j=0; j<n; j++)
			for(k=0; k<n; k++)
	{
		v[0]=mins[0]+d[0]*i;
		v[1]=mins[1]+d[1]*j;
		v[2]=mins[2]+d[2]*k;
		BSDE_PlaneD_TransformPoint(v, pos, pts+(i*n2+j*n+k)*3);
	}
}

int BSDE_SolidOBB_ClipPoints(
	double *mins, double *maxs, double *pos,
	double *ipts, double *opts, int npts)
{
	int i, j, k;

	k=0;
	for(i=0; i<npts; i++)
	{
		j=BSDE_SolidOBB_ObbPointCollide(
			mins, maxs, pos, ipts+i*3);

		if(j)
		{
			V3D_COPY(ipts+i*3, opts+k*3);
			k++;
		}
	}
	return(k);
}

int BSDE_SolidOBB_ObbCollidePoints(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *pts)
{
	double apts[8*3], bpts[8*3];
	int i, n;

	for(i=0; i<8; i++)
	{
		V3D_SCALE(apos+0*4, (i&1)?1:-1, apts+i*3);
		V3D_ADDSCALE(apts+i*3, apos+1*4, (i&2)?1:-1, apts+i*3);
		V3D_ADDSCALE(apts+i*3, apos+2*4, (i&4)?1:-1, apts+i*3);
	}
}

#if 1
int BSDE_SolidOBB_CollideContact(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double pts[8*3];
	double ext[4];
	double bn[3], bo[3], v0[3], v1[3];
	double bd, d0, d1;
	int i, j, k;

	i=BSDE_SolidOBB_CheckCollide(
		amins, amaxs, apos, bmins, bmaxs, bpos);
	if(!i)return(0);

	V3D_ZERO(bn);
	V3D_ZERO(bo);

	BSDE_SolidOBB_GenPoints(amins, amaxs, apos, pts);
	i=BSDE_SolidOBB_ClipPoints(bmins, bmaxs, bpos, pts, pts, 8);

	for(j=0; j<i; j++)
	{
		V3D_ADD(bo, pts+j*3, bo);

		BSDE_SolidOBB_ObbBestNormalPoint(
			bmins, bmaxs, bpos, pts+j*3, v0, &d0);
		V3D_ADD(bn, v0, bn);

//		if(d0<0.001)d0=0.001;
//		V3D_ADDSCALE(bn, v0, 1.0/d0, bn);
	}

	BSDE_SolidOBB_GenPoints(bmins, bmaxs, bpos, pts);
	k=BSDE_SolidOBB_ClipPoints(amins, amaxs, apos, pts, pts, 8);

	for(j=0; j<k; j++)
	{
		V3D_ADD(bo, pts+j*3, bo);

		BSDE_SolidOBB_ObbBestNormalPoint(
			amins, amaxs, apos, pts+j*3, v0, &d0);
		V3D_ADDSCALE(bn, v0, -1, bn);

//		if(d0<0.001)d0=0.001;
//		V3D_ADDSCALE(bn, v0, -1.0/d0, bn);
	}

//	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, bn, ext);
//	d0=ext[1]-ext[0];

	i=i+k;
	if(i<1)
	{
		BSDE_SolidOBB_ContactPoint(
			amins, amaxs, apos, bmins, bmaxs, bpos, org);
		BSDE_SolidOBB_ObbBestNormalPoint(
			bmins, bmaxs, bpos, org, bn, &d0);

		V3D_NORMALIZE(bn, norm);
		*dist=d0;

		return(1);
	}


	if(0)
	{
		V3D_SCALE(bo, 1.0/i, org);
		BSDE_SolidOBB_ObbBestNormalPoint(
			amins, amaxs, apos, org, v0, &d0);
		BSDE_SolidOBB_ObbBestNormalPoint(
			bmins, bmaxs, bpos, org, v1, &d1);
		V3D_ADDSCALE(v1, v0, -1, bn);
	}

	if(0)
	{
		V3D_SCALE(bo, 1.0/i, org);
		BSDE_SolidOBB_ObbBestNormalPoint(
			amins, amaxs, apos, org, v0, &d0);
		BSDE_SolidOBB_ObbBestNormalPoint(
			bmins, bmaxs, bpos, org, v1, &d1);

#if 0
		i=BSDE_SolidOBB_ObbBestNormalObb(
			amins, amaxs, apos, bmins, bmaxs, bpos, v0, &d0);

		j=BSDE_SolidOBB_ObbBestNormalObb(
			bmins, bmaxs, bpos, amins, amaxs, apos, v1, &d1);
#endif

		if(d0<d1)
		{
			V3D_SCALE(v0, -1, bn);
			bd=d0;
		}else
		{
			V3D_COPY(v1, bn);
			bd=d1;
		}

		V3D_COPY(bn, norm);
		*dist=bd;

		return(1);
	}


	V3D_NORMALIZE(bn, bn);

	BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, bn, ext+0);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, bn, ext+2);
	d0=(ext[0]>ext[2])?ext[0]:ext[2];
	d1=(ext[1]<ext[3])?ext[1]:ext[3];
	bd=d1-d0;

	if(bd>0.1)bd=0.1;

	V3D_SCALE(bo, 1.0/i, org);

	V3D_COPY(bn, norm);
	*dist=bd;

	return(1);
}
#endif


int BSDE_SolidOBB_CollideMultiContact(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	BSDE_World *world, BSDE_Solid *aobj, BSDE_Solid *bobj,
	BSDE_Contact **lst)
{
	double bn[3], rbn[3], v0[3], v1[3];
	double ext[4];
	double dv[3], aorg[3], borg[3];
	double org[3];
	double d0, d1;
	int i, j, n;

	i=BSDE_SolidOBB_CheckCollide(
		amins, amaxs, apos, bmins, bmaxs, bpos);
	if(!i)return(0);

	BSDE_SolidOBB_ContactPoint(
		amins, amaxs, apos, bmins, bmaxs, bpos, org);

	n=0;

	for(i=0; i<3; i++)
	{
		V3D_COPY(apos+(i*4), bn);
		V3D_SCALE(bn, -1, rbn);

		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bn, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bn, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		if((ext[2]>=ext[0]) && (ext[2]<=ext[1]))
		{
			lst[n++]=BSDE_MakeContact(world, bobj, aobj,
				org, bn, d1-d0);
		}

		if((ext[3]>=ext[0]) && (ext[3]<=ext[1]))
		{
			lst[n++]=BSDE_MakeContact(world, bobj, aobj,
				org, rbn, d1-d0);
		}

#if 1
		V3D_COPY(bpos+(i*4), bn);
		V3D_SCALE(bn, -1, rbn);

		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bn, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bn, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		if((ext[2]>=ext[0]) && (ext[2]<=ext[1]))
		{
			lst[n++]=BSDE_MakeContact(world, aobj, bobj,
				org, bn, d1-d0);
		}

		if((ext[3]>=ext[0]) && (ext[3]<=ext[1]))
		{
			lst[n++]=BSDE_MakeContact(world, aobj, bobj,
				org, rbn, d1-d0);
		}
#endif
	}
	return(n);
}
