///////////////////////////////////////////////////////////////////////////////
/*
Assert:
There exists a single point where a collision can be resolved.
The best normal is the axis of least penetration.
The best point is the point of most penetration.

First, figure axis.
Gen points.
Look for best point.
Otherwise, try to come up with a point (generating a set from edge clipping).
 */

#include "bsde_internal.h"

int BSDE_SolidOBB3_CollideNormal(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *norm, double *dist)
{
	double ext[4];
	double bn[4], v0[3], v1[3];
	double d0, d1, d, bd, f;
	int i, j, k;

	V3D_ZERO(bn);
	bd=99999999.0;

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			apos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			apos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);


		j=(V3D_DOT(apos+3*4, apos+i*4) >
			V3D_DOT(bpos+3*4, apos+i*4));
		if(d<bd)
		{
			V3D_SCALE(apos+i*4, j?1:-1, bn);
			bn[4]=j?amaxs[i]:amins[i];
			bd=d;
		}
	}

	for(i=0; i<3; i++)
	{
		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			bpos+i*4, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);


		j=(V3D_DOT(apos+3*4, bpos+i*4) >
			V3D_DOT(bpos+3*4, bpos+i*4));
		if(d<bd)
		{
			V3D_SCALE(bpos+i*4, j?1:-1, bn);
			bn[4]=j?bmaxs[i]:bmins[i];
			bd=d;
		}
	}

#if 1
	for(i=0; i<3; i++)
		for(j=0; j<3; j++)
	{
		V3D_CROSS(apos+i*4, bpos+j*4, v0);
		V3D_NORMALIZE(v0, v0);

		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, v0, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, v0, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);

		k=(V3D_DOT(apos+3*4, v0) >
			V3D_DOT(bpos+3*4, v0));
		if(d<bd)
		{
			V3D_SCALE(v0, k?1:-1, bn);
			bn[4]=0; //not really a plane
			bd=d;
		}
	}
#endif

	V4D_COPY(bn, norm);
	*dist=bd;
	return(1);
}

void BSDE_SolidOBB3_GenPoints(
	double *mins, double *maxs, double *pos, double *pts)
{
	double v[3];
	int i, n;

	for(i=0; i<8; i++)
	{
		v[0]=(i&1)?maxs[0]:mins[0];
		v[1]=(i&2)?maxs[1]:mins[1];
		v[2]=(i&4)?maxs[2]:mins[2];
		BSDE_PlaneD_TransformPoint(v, pos, pts+i*3);
	}
}

void BSDE_SolidOBB3_ObbBestNormalPoint(
	double *mins, double *maxs, double *pos,
	double *org, double *norm)
{
	double bn[3];
	double n[3], o[3];
	double d, f, bd;
	int i;

	V3D_COPY(pos+12, o);

	V3D_ZERO(bn);
	bd=99999999.0;

	for(i=0; i<3; i++)
	{
		V3D_COPY(pos+(i*4), n);
		d=V3D_DOT(org, n)-V3D_DOT(o, n);

		f=d-mins[i];
		if(f<bd)
		{
			V3D_SCALE(n, -1, bn);
			bd=f;
		}

		f=maxs[i]-d;
		if(f<bd)
		{
			V3D_SCALE(n, 1, bn);
			bd=f;
		}
	}

	V3D_NORMALIZE(bn, norm);
}

int BSDE_SolidOBB3_CheckPointCollide(
	double *mins, double *maxs, double *pos, double *pt)
{
	double r[3], v[3];

	V3D_SUB(pt, pos+3*4, r);
	v[0]=V3D_DOT(r, pos+0*3);
	v[1]=V3D_DOT(r, pos+1*3);
	v[2]=V3D_DOT(r, pos+2*3);
	if((v[0]<mins[0]) || (v[0]>maxs[0]))return(0);
	if((v[1]<mins[1]) || (v[1]>maxs[1]))return(0);
	if((v[2]<mins[2]) || (v[2]>maxs[2]))return(0);
	return(1);
}

int BSDE_SolidOBB3_CollideContact(
	double *amins, double *amaxs, double *apos,
	double *bmins, double *bmaxs, double *bpos,
	double *rorg, double *rnorm, double *rdist)
{
	static int edges[24]={
	0, 1,  1, 2,  2, 3,  3, 0,
	4, 5,  5, 6,  6, 7,  7, 4,
	4, 0,  5, 1,  6, 2,  7, 3};

//	double apts[64*3], bpts[64*3];
	double apts[512*3], bpts[512*3];
	double ext[4];
	double norm[4], ndist;
	double bp[3], bd, d, d0, d1;
	double *v0, *v1, f, g, tv0[3], tv1[3];
	int i, j, k;

	i=BSDE_SolidOBB3_CollideNormal(amins, amaxs, apos,
		bmins, bmaxs, bpos, norm, &ndist);
	if(!i)return(0);

#if 0
	BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, norm, ext+0);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, norm, ext+2);
	d0=(ext[0]>ext[2])?ext[0]:ext[2];
	d1=(ext[1]<ext[3])?ext[1]:ext[3];

//	BSDE_SolidOBB3_GenPoints(amins, amaxs, apos, apts);
//	BSDE_SolidOBB3_GenPoints(bmins, bmaxs, bpos, bpts);

	BSDE_SolidOBB_GenPointsNx(amins, amaxs, apos, apts, 8);
	BSDE_SolidOBB_GenPointsNx(bmins, bmaxs, bpos, bpts, 8);

	bd=-1;
	k=0;
	V3D_ZERO(bp);
#endif

#if 0
	for(i=0; i<512; i++)
	{
		j=BSDE_SolidOBB3_CheckPointCollide(
			amins, amaxs, apos, bpts+i*3);
		if(!j)continue;

		V3D_ADD(bp, bpts+i*3, bp);
		k++;
	}

	for(i=0; i<512; i++)
	{
		j=BSDE_SolidOBB3_CheckPointCollide(
			bmins, bmaxs, bpos, apts+i*3);
		if(!j)continue;

		V3D_ADD(bp, apts+i*3, bp);
		k++;
	}

	if(k)
	{
		//ok, use this
		V3D_SCALE(bp, 1.0/k, rorg);

//		V3D_SUB(apos+3*4, bp, tv0);
//		V3D_SUB(bp, bpos+3*4, tv1);


		V3D_NORMALIZE(tv0, tv0);
		V3D_NORMALIZE(tv1, tv1);

		BSDE_SolidOBB3_ObbBestNormalPoint(
			amins, amaxs, apos, rorg, tv0);
		BSDE_SolidOBB3_ObbBestNormalPoint(
			amins, amaxs, apos, rorg, tv1);

		V3D_ADDSCALE(tv1, tv0, -1, rnorm);
		V3D_NORMALIZE(rnorm, rnorm);

//		V3D_SCALE(norm, 1, rnorm);

		BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos,
			rnorm, ext+0);
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			rnorm, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		*rdist=d1-d0;
		return(1);

//		V3D_SCALE(norm, 1, rnorm);
//		*rdist=ndist;
//		return(1);
	}

	return(0);
#endif

	//else, generate a push in the general direction
	V3D_SUB(apos+3*4, bpos+3*4, norm);
	V3D_NORMALIZE(norm, norm);

	BSDE_SolidOBB_ObbPlaneExtents(amins, amaxs, apos, norm, ext+0);
	BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos, norm, ext+2);
	d0=(ext[0]>ext[2])?ext[0]:ext[2];
	d1=(ext[1]<ext[3])?ext[1]:ext[3];
	if(d0>=d1)return(0);	//somehow, no collision

	d=(d0+d1)/2-V3D_DOT(bpos+3*4, norm);
	V3D_ADDSCALE(bpos+3*4, norm, d, rorg);
	V3D_COPY(norm, rnorm);
	*rdist=d1-d0;
	return(1);
}
