#include "bsde_internal.h"

int BSDE_SolidSphere_CollideContact(
	double arad, double *aorg, double brad, double *borg,
	double *org, double *norm, double *dist)
{
	double dv[3];
	double l, f;

	V3D_SUB(aorg, borg, dv);
	l=V3D_NORMALIZE(dv, dv);
	if(l>=(arad+brad))return(0);

//	l=V3D_DOT(dv, dv);
//	f=arad*arad+brad*brad;
//	if(l>=f)return(0);

	l=V3D_NORMALIZE(dv, dv);

	if(org)
	{
		f=(brad+(l-arad))*0.5;
		if(f<0)f=0;

		V3D_ADDSCALE(borg, dv, f, org);
		V3D_COPY(dv, norm);
		*dist=(arad+brad)-l;
	}

	return(1);
}

#if 0
int BSDE_SolidSphere_CollideObbContact(
	double arad, double *aorg,
	double *bmins, double *bmaxs, double *pos,
	double *org, double *norm, double *dist)
{
	double pt[3], dv[3];
	double d;

	BSDE_SolidOBB_ObbNearestPoint(bmins, bmaxs, pos, aorg, pt);
	V3D_SUB(aorg, pt, dv);
	d=V3D_NORMALIZE(dv, dv);
	if(d<0.0001)
	{
		dv[2]=1;
		pt[2]-=arad;
	}

	if(d>=arad)return(0);

	if(org)
	{
		V3D_COPY(pt, org);
		V3D_COPY(dv, norm);
		*dist=arad-d;
	}

	return(1);
}
#endif

int BSDE_SolidSphere_CollideObbContact(
	double arad, double *aorg,
	double *bmins, double *bmaxs, double *bpos,
	double *org, double *norm, double *dist)
{
	double pt[3], dv[3];
	double ext[4];
	double bn[3];
	double d0, d1, d, bd, bo;
	int i, j;

	BSDE_SolidOBB_ObbNearestPoint(bmins, bmaxs, bpos, aorg, pt);
	V3D_SUB(aorg, pt, dv);
	d=V3D_NORMALIZE(dv, dv);
	if(d<0.0001)
	{
		dv[2]=1;
		pt[2]-=arad;
	}

	if(d>=arad)return(0);

	V3D_COPY(dv, bn);
	bd=arad-d;

	for(i=0; i<3; i++)
	{
		ext[0]=V3D_DOT(aorg, bpos+i*4)-arad;
		ext[1]=V3D_DOT(aorg, bpos+i*4)+arad;
		BSDE_SolidOBB_ObbPlaneExtents(bmins, bmaxs, bpos,
			bpos+i*4, ext+2);
		d0=(ext[0]>ext[2])?ext[0]:ext[2];
		d1=(ext[1]<ext[3])?ext[1]:ext[3];

		d=d1-d0;
		if(d<0)return(0);

		if(d<bd)
		{
			V3D_COPY(bpos+i*4, bn);
			bd=d;
		}
	}

	if(org)
	{
		V3D_COPY(pt, org);
		V3D_NORMALIZE(bn, bn);
		V3D_COPY(bn, norm);
		*dist=bd;
	}
	return(1);
}
